"""train_utils.py
=================
Factory helpers to wire **datasets**, **model + optim + scheduler**, **loss_fn**
and the `TransformerUnfreezeScheduler` that cooperates with the *DepthTrainer*
(re-written 2025-05-09).

Key features
------------
* Zero duplicate class definitions.
* Cosine-Annealing-WarmRestarts scheduler contado em **micro-steps**.
* Albumentations wrapper with NaN-mask fix, *picklable* (multiprocess-safe).
* Boom-log: every unfreeze event appended to ``unfreeze_log.jsonl``.
* Fully-typed public API:
    ─ ``build_dataloaders``
    ─ ``build_model_optim_sched``
    ─ ``build_loss_fn``
    ─ ``search_best_scale``
    ─ ``TransformerUnfreezeScheduler``
"""
from __future__ import annotations

from pathlib import Path
from typing import Dict, List, Tuple

import json, time
import numpy as np
import torch
from torch.utils.data import DataLoader
from torch.optim import AdamW
from torch.optim.lr_scheduler import CosineAnnealingWarmRestarts

# project-internal ————————————————————————————————————————————————
from panorai.depth import ModelRegistry
from panorai.depth.trainers.losses import FixedDepthLoss
from panorai.depth.trainers.transforms import PrepareForNet
from panorai.depth.trainers.estimate_scales import estimate_best_scale
from custom_data import load_datasets as _load_datasets, get_cypher, collate_fn
from custom_data.disk_cached_transform import DiskCachedTransform
from torchvision.transforms import Compose

# ════════════════════════════════════════════════════════════════════════════
# Albumentations: global pipelines  →  picklable
# ════════════════════════════════════════════════════════════════════════════
import os
os.environ["NO_ALBUMENTATIONS_UPDATE"] = "1"
import albumentations as A

AUG_TRAIN = A.Compose([
    # A.RandomBrightnessContrast(brightness_limit=0.1, contrast_limit=0.2, p=0.5),
    # A.GaussianBlur(blur_limit=3, p=0.2),


    A.CoarseDropout(
            num_holes_range=[1, 5],
            hole_height_range=[0.1, 0.2],
            hole_width_range=[0.1, 0.2],
            fill_mask=0.0001,
            fill=0,
            p=0.5
        ),
    A.Normalize(mean=[0.485, 0.456, 0.406],
                std=[0.229, 0.224, 0.225]),
])

AUG_VAL = A.Compose([
    A.Normalize(mean=[0.485, 0.456, 0.406],
                std=[0.229, 0.224, 0.225]),
])


# class AlbumentationsWrapper:
#     """Callable que aplica Albumentations e corrige *NaNs* na depth."""

#     def __init__(self, aug: A.Compose):
#         self.aug = aug

#     def __call__(self, sample: Dict[str, np.ndarray]) -> Dict[str, np.ndarray]:
#         img = sample["rgb_image"].astype(np.uint8)
#         depth = sample["xyz_image"]

#         img_out, depth_out = [], []
#         for i in range(img.shape[0]):
#             res = self.aug(image=np.ascontiguousarray(img[i]))
#             aug_img = res["image"]

#             mask = np.all(aug_img == 0, axis=-1)
#             d = depth[i].copy()
#             d[mask] = 1e-6  # evita NaNs na loss

#             img_out.append(aug_img)
#             depth_out.append(d)

#         sample["rgb_image"] = np.stack(img_out, 0)
#         sample["xyz_image"] = np.stack(depth_out, 0)
#         return sample
import cv2
import numpy as np
import matplotlib.pyplot as plt

EDGE_DEBUG_MAX_PLOTS = 5
edge_debug_counter = [0]  # mutable global counter

class AlbumentationsWrapper:
    """Callable que aplica Albumentations, corrige *NaNs* e remove bordas ruidosas na profundidade."""

    def __init__(self, aug: A.Compose, edge_ksize: int = 3, edge_thresh: float = 10.0, debug: bool = False):
        self.aug = aug
        self.edge_ksize = edge_ksize
        self.edge_thresh = edge_thresh
        self.debug = debug

    def __call__(self, sample: Dict[str, np.ndarray]) -> Dict[str, np.ndarray]:
        img = sample["rgb_image"].astype(np.uint8)
        depth = sample["xyz_image"]

        img_out, depth_out = [], []
        for i in range(img.shape[0]):
            res = self.aug(image=np.ascontiguousarray(img[i]))
            aug_img = res["image"]

            mask = np.all(aug_img == 0, axis=-1)
            d = depth[i].copy()
            d[mask] = 1e-6  # evita NaNs na loss

            img_out.append(aug_img)
            depth_out.append(d)

        sample["rgb_image"] = np.stack(img_out, 0)
        sample["xyz_image"] = np.stack(depth_out, 0)

        # # # ───────────────────────
        # # # Mask strong Z-gradients
        # # # ───────────────────────
        # # masked = []
        # # for d in sample["xyz_image"]:
        # #     if d.ndim != 3 or d.shape[2] < 3:
                
        # #         masked.append(d)
        # #         continue

        # #     z = d[:, :, 2].astype(np.float32)
        # #     gx = cv2.Sobel(z, cv2.CV_32F, 1, 0, ksize=self.edge_ksize)
        # #     gy = cv2.Sobel(z, cv2.CV_32F, 0, 1, ksize=self.edge_ksize)
        # #     gmag = np.sqrt(gx**2 + gy**2)
        # #     mask = (gmag < self.edge_thresh).astype(np.float32)

        # #     d_out = d.copy()
        # #     d_out[:, :, 2] *= mask
        # #     masked.append(d_out)

        # #     # ─── Debug Plot ─────────────────────────
        # #     if self.debug and edge_debug_counter[0] < EDGE_DEBUG_MAX_PLOTS:
        # #         edge_debug_counter[0] += 1
        # #         fig, axs = plt.subplots(1, 3, figsize=(12, 4))
        # #         axs[0].imshow(z, cmap='viridis')
        # #         axs[0].set_title("Original Z-Depth")
        # #         axs[1].imshow(gmag, cmap='inferno')
        # #         axs[1].set_title("Z Gradient Magnitude")
        # #         axs[2].imshow(z * mask, cmap='viridis')
        # #         axs[2].set_title("Masked Z (after filter)")
        # #         for ax in axs:
        # #             ax.axis('off')
        # #         plt.tight_layout()
        # #         plt.show()

        # sample["xyz_image"] = np.stack(masked, 0)
        return sample

def _get_aug(eval_mode: bool = False) -> AlbumentationsWrapper:
    """Devolve wrapper picklável (usado pelos *DataLoader* workers)."""
    return AlbumentationsWrapper(AUG_VAL if eval_mode else AUG_TRAIN)

# # ════════════════════════════════════════════════════════════════════════════
# # Data loaders
# # ════════════════════════════════════════════════════════════════════════════
# def build_dataloaders(cfg: Dict) -> Tuple[DataLoader, DataLoader]:
#     ds = cfg["dataset"]
#     size = ds.get("size", 518)
#     n_tr, n_val = ds.get("n_angles_train", 10), ds.get("n_angles_val", 1)
#     bs = ds.get("batch_size", 2)

#     tr_tf = DiskCachedTransform(
#         transform_fn=Compose([
#             PrepareForNet(size=size, n_angles=n_tr, max_angle_deg=45, eval_mode=False),
#         ]),
#         post_transform_fn=_get_aug(eval_mode=False),
#         cache_dir=ds.get("train_cache_dir", ".cache/train"),
#     )
#     val_tf = DiskCachedTransform(
#         transform_fn=Compose([
#             PrepareForNet(size=size, n_angles=n_val, max_angle_deg=45, eval_mode=True),
#         ]),
#         post_transform_fn=_get_aug(eval_mode=True),
#         cache_dir=ds.get("val_cache_dir", ".cache/val"),
#     )

#     datasets = _load_datasets(
#         ds["train_name"], ds["val_name"], get_cypher(),
#         train_transform=tr_tf, valid_transform=val_tf,
#         n_angles_train=n_tr, n_angles_val=n_val,
#     )

#     tr_loader = DataLoader(
#         datasets["trainset"], batch_size=bs, shuffle=True, drop_last=True,
#         num_workers=ds.get("train_workers", 3), prefetch_factor=2,
#         collate_fn=collate_fn, persistent_workers=True,
#     )
#     val_loader = DataLoader(
#         datasets["valset"], batch_size=1, shuffle=False, drop_last=False,
#         num_workers=ds.get("val_workers", 4), prefetch_factor=2,
#         collate_fn=collate_fn, persistent_workers=True,
#     )
#     return tr_loader, val_loader

# ─────────────────────────────────────────────────────────────────────────────
# Data loaders
# ─────────────────────────────────────────────────────────────────────────────
from custom_data.disk_cached_transform import TeacherCacher

# def build_dataloaders(cfg: Dict,
#                       teacher_model: torch.nn.Module | None = None
#                      ) -> Tuple[DataLoader, DataLoader]:
#     """
#     If `teacher_model` is passed, the train transform will call / cache
#     its coarse predictions.  Validation stays teacher-free.
#     """
#     ds   = cfg["dataset"]
#     size = ds.get("size", 518)
#     n_tr, n_val = ds.get("n_angles_train", 10), ds.get("n_angles_val", 1)
#     bs   = ds.get("batch_size", 2)

#     # ─── optional teacher cacher ---------------------------------------
#     teacher_cacher = None
#     if teacher_model is not None:
#         teacher_cacher = TeacherCacher(
#             cfg,
#             cache_root=ds.get("teacher_cache_dir", ".cache/teacher"),
#             device = cfg.get("device", "mps")
#         )

#     # # ─── train transform (with teacher) --------------------------------
#     # tr_tf = DiskCachedTransform(
#     #     transform_fn=Compose([
#     #         PrepareForNet(size=size, n_angles=n_tr, max_angle_deg=45),
#     #     ]),
#     #     post_transform_fn=_get_aug(eval_mode=False),
#     #     cache_dir=ds.get("train_cache_dir", ".cache/train"),
#     #     teacher_cacher=teacher_cacher,          # ← inject here
#     #     train_mode=True,
#     #     n_angles=n_tr,
#     #     max_angle_deg=45
#     # )

#     # # ─── validation transform (no teacher) -----------------------------
#     # val_tf = DiskCachedTransform(
#     #     transform_fn=Compose([
#     #         PrepareForNet(size=size, n_angles=n_val, max_angle_deg=45),
#     #     ]),
#     #     post_transform_fn=_get_aug(eval_mode=True),
#     #     cache_dir=ds.get("val_cache_dir", ".cache/val"),
#     #     teacher_cacher=teacher_cacher,                     # keep val clean
#     #     train_mode=False,
#     #     n_angles=n_val,
#     #     max_angle_deg=45
#     # )

#     # ─── train transform (with teacher) --------------------------------
#     tr_tf = DiskCachedTransform(
#         transform_fn=Compose([
#             PrepareForNet(size=size, n_angles=n_tr, max_angle_deg=45),
#         ]),
#         post_transform_fn=_get_aug(eval_mode=False),
#         cache_dir=ds.get("train_cache_dir", ".cache/train"),
#         teacher_cacher=teacher_cacher,          # teacher only in training
#         train_mode=True,
#         n_angles=n_tr,
#         max_angle_deg=45,
#         # only cache two of your four angles (for example)
#         #cache_angle_idxs=[0, 2],
#     )

#     # ─── validation transform (no teacher) -----------------------------
#     val_tf = DiskCachedTransform(
#         transform_fn=Compose([
#             PrepareForNet(size=size, n_angles=n_val, max_angle_deg=45),
#         ]),
#         post_transform_fn=_get_aug(eval_mode=True),
#         cache_dir=ds.get("val_cache_dir", ".cache/val"),
#         teacher_cacher=None,                    # no teacher at validation
#         train_mode=False,
#         n_angles=n_val,
#         max_angle_deg=45,
#         # for val we only ever need the zero‐angle, so cache that alone
#         cache_angle_idxs=[0],
#     )

#     # ─── dataset loading ----------------------------------------------
#     datasets = _load_datasets(
#         ds["train_name"], ds["val_name"], get_cypher(),
#         train_transform=tr_tf, valid_transform=val_tf,
#         n_angles_train=n_tr, n_angles_val=n_val,
#     )

#     # ─── data loaders --------------------------------------------------
#     tr_loader = DataLoader(
#         datasets["trainset"], batch_size=bs, shuffle=True, drop_last=True,
#         num_workers=ds.get("train_workers", 3), prefetch_factor=2,
#         collate_fn=collate_fn, persistent_workers=True,
#     )
#     val_loader = DataLoader(
#         datasets["valset"], batch_size=1, shuffle=False, drop_last=False,
#         num_workers=ds.get("val_workers", 4), prefetch_factor=2,
#         collate_fn=collate_fn, persistent_workers=True,
#     )
#     return tr_loader, val_loader

from custom_data.disk_cached_transform import TeacherCacher, DiskCachedTransform, LMDBCachedTransform

def build_dataloaders(cfg: Dict,
                      teacher_model: torch.nn.Module | None = None
                     ) -> Tuple[DataLoader, DataLoader]:
    """
    If `teacher_model` is passed, the train transform will use LMDB caching
    for expensive projections via LMDBCachedTransform. Validation remains teacher-free.
    """
    ds   = cfg["dataset"]
    size = ds.get("size", 518)
    n_tr, n_val = ds.get("n_angles_train", 10), ds.get("n_angles_val", 1)
    bs   = ds.get("batch_size", 2)

    # Optional teacher cacher
    teacher_cacher = None
    if teacher_model is not None:
        teacher_cacher = TeacherCacher(
            cfg,
            cache_root=ds.get("teacher_cache_dir", ".cache/teacher"),
            device=cfg.get("device", "mps")
        )

    # Train transform with LMDB cache
    tr_tf = LMDBCachedTransform(
        transform_fn=Compose([
            PrepareForNet(size=size, n_angles=n_tr, max_angle_deg=45),
        ]),
        cache_dir=ds.get("train_cache_dir", ".cache/train"),
        lmdb_dir=ds.get("train_lmdb_dir", None),
        post_transform_fn=_get_aug(eval_mode=False),
        teacher_cacher=teacher_cacher,
        train_mode=True,
        seed=cfg.get("seed", 42),
        n_angles=n_tr,
        max_angle_deg=45,
        cache_angle_idxs=ds.get("cache_angle_idxs_train", None),
    )

    # Validation transform without teacher, optional LMDB
    val_tf = LMDBCachedTransform(
        transform_fn=Compose([
            PrepareForNet(size=size, n_angles=n_val, max_angle_deg=45),
        ]),
        cache_dir=ds.get("val_cache_dir", ".cache/val"),
        lmdb_dir=ds.get("val_lmdb_dir", None),
        post_transform_fn=_get_aug(eval_mode=True),
        teacher_cacher=None,
        train_mode=False,
        seed=cfg.get("seed", 42),
        n_angles=n_val,
        max_angle_deg=45,
        cache_angle_idxs=ds.get("cache_angle_idxs_val", [0]),
    )

    # Dataset loading
    datasets = _load_datasets(
        ds["train_name"], ds["val_name"], get_cypher(),
        train_transform=tr_tf, valid_transform=val_tf,
        n_angles_train=n_tr, n_angles_val=n_val,
    )

    # DataLoaders
    tr_loader = DataLoader(
        datasets["trainset"], batch_size=bs, shuffle=True, drop_last=True, pin_memory=True,
        num_workers=ds.get("train_workers", 3), prefetch_factor=2,
        collate_fn=collate_fn, persistent_workers=True,
    )
    val_loader = DataLoader(
        datasets["valset"], batch_size=1, shuffle=False, drop_last=False,pin_memory=True,
        num_workers=ds.get("val_workers", 4), prefetch_factor=1,
        collate_fn=collate_fn, persistent_workers=True,
    )
    return tr_loader, val_loader


# ════════════════════════════════════════════════════════════════════════════
# Model + optimiser + scheduler
# ════════════════════════════════════════════════════════════════════════════
# In train_utils.py, replace your existing _freeze_initial with this:

# def _freeze_initial(model: torch.nn.Module):
#     # 1) Freeze everything
#     for p in model.parameters():
#         p.requires_grad = False

#     # 2) Unfreeze entire depth head only
#     for name, p in model.named_parameters():
#         if name.startswith("depth_head.scratch."):
#             p.requires_grad = True

def _freeze_initial(model: torch.nn.Module):
    # 1) Freeze everything
    for p in model.parameters():
        p.requires_grad = False

    # 2) Unfreeze all of depth head
    for name, p in model.named_parameters():
        if name.startswith("depth_head.scratch."):
            p.requires_grad = True

    # 3) Unfreeze global layer norm (used for scaling transformer outputs)
    for name, p in model.named_parameters():
        if name.startswith("pretrained.norm.weight"):
            p.requires_grad = True
import re
import torch
import re
import torch

def _freeze_initial(
    model: torch.nn.Module,
    *,
    min_block: int = 20
):
    """
    Freeze everything except:
      • your core depth_head outputs
      • LayerNorm weight & bias in backbone blocks >= min_block
      • the final pretrained.norm
      • (you can still tweak patch/embed etc. if you like)
    """
    named = dict(model.named_parameters())

    # 1) freeze all
    for p in named.values():
        p.requires_grad = False

    # 2) depth_head outputs
    head_allow = (
        "depth_head.projects.",
        "depth_head.scratch.output_conv1",
        "depth_head.scratch.output_conv2.0",
        "depth_head.scratch.output_conv2.2",
        "depth_head.scratch.head_logstd",
        "depth_head.scratch.head_masklog",
    )
    for name, p in named.items():
        if any(name.startswith(pref) for pref in head_allow):
            p.requires_grad = True

    # 3) LayerNorm in blocks >= min_block
    block_norm_re = re.compile(
        r"^pretrained\.blocks\.(\d+)\.norm[12]\.(weight|bias)$"
    )
    for name, p in named.items():
        m = block_norm_re.match(name)
        if m and int(m.group(1)) >= min_block:
            p.requires_grad = True

    # 4) final pretrained.norm
    for suffix in ("weight", "bias"):
        key = f"pretrained.norm.{suffix}"
        if key in named:
            named[key].requires_grad = True

    # # 6) Unfreeze patch embedding & positional embeddings
    # for emb in ("patch_embed.proj.weight",
    #             "patch_embed.proj.bias",
    #             "pos_embed",
    #             "cls_token"):
    #     key = f"pretrained.{emb}"
    #     if key in named:
    #         named[key].requires_grad = True

     # ── graph cut: detach outputs of blocks 0..min_block-1 ─────────────


# def _layerwise_lr(model: torch.nn.Module, base_lr: float, decay: float = 0.95):
#     """
#     Assign higher learning rates to head layers and norm for fast adaptation,
#     while decaying backbone block rates.
#     """
#     groups = []
#     for name, p in model.named_parameters():
#         if not p.requires_grad:
#             continue

#         # LayerNorm: allow rapid recalibration
#         if name.startswith("pretrained.norm."):
#             lr = base_lr * 2.0

#         # Depth head, final conv (output_conv2): most aggressive
#         elif name.startswith("depth_head.scratch.output_conv2"):
#             lr = base_lr * 5.0

#         # Depth head, first conv (output_conv1): strong but less than last layer
#         elif name.startswith("depth_head.scratch.output_conv1"):
#             lr = base_lr * 3.0

#         # Any other head parameters
#         elif name.startswith("depth_head."):
#             lr = base_lr * 2.0

#         # Backbone blocks: layer‐wise decay from start_block down
#         elif name.startswith("pretrained.blocks."):
#             blk = int(name.split(".")[2])
#             lr = base_lr * (decay ** (23 - blk))

#         # Bias terms: usually lower rate
#         elif name.endswith(".bias"):
#             lr = base_lr * 0.5

#         # Default
#         else:
#             lr = base_lr

#         groups.append({"params": [p], "lr": lr})

#     return groups

def _layerwise_lr(model: torch.nn.Module, base_lr: float,
                  decay: float = 0.95,
                  refiner_scale: float = 1.0):      # ← novo arg
    groups = []
    for name, p in model.named_parameters():
        if not p.requires_grad:
            continue

        # ─── 1. refiner primeiro ─────────────────────────────
        if name.startswith("depth_head.refiner."):
            lr = base_lr * refiner_scale             # <<< aqui
        # ─── 2. regras existentes ────────────────────────────
        elif name.startswith("pretrained.norm."):
            lr = base_lr * 2.0
        elif name.startswith("depth_head.scratch.output_conv2"):
            lr = base_lr * 5.0
        elif name.startswith("depth_head.scratch.output_conv1"):
            lr = base_lr * 3.0
        elif name.startswith("depth_head."):
            lr = base_lr * 2.0
        elif name.startswith("pretrained.blocks."):
            blk = int(name.split(".")[2])
            lr = base_lr * (decay ** (23 - blk))
        elif name.endswith(".bias"):
            lr = base_lr * 0.5
        else:
            lr = base_lr

        groups.append({"params": [p], "lr": lr})

    return groups

class ResettableScheduler:
    """Cosine-annealing + warm restarts in *micro-steps*."""
    def __init__(self, optim: AdamW, cfg: Dict, steps_per_epoch: int, resetable: bool = True):
        self.resetable = resetable
        self.optim = optim
        self.cfg = cfg
        self.spe = max(1, steps_per_epoch)
        self._microstep = 0
        self._make_sched()

    def _make_sched(self):
        T_0 = self.cfg.get("sgdr_restart_period", 30) * self.spe
        self.base = CosineAnnealingWarmRestarts(
            self.optim, T_0=T_0,
            T_mult=self.cfg.get("sgdr_restart_mult", 2),
            eta_min=float(self.cfg.get("min_lr", 1e-7)),
        )

    def step(self):
        self._microstep += 1
        self.base.step(self._microstep)  # Important!

    def reset(self):
        if self.resetable:
            self._make_sched()
            self._microstep = 0


import warnings, torch

# ─── compile helper ──────────────────────────────────────────────────────────
import warnings, torch

# panorai_models/utils/compile_helper.py
import torch, warnings, os


def maybe_compile(model, *, enabled=True, mode="reduce-overhead", backend="aot_eager"):
    """
    Compila se estiver habilitado; em caso de erro (ou backend MPS instável)
    volta para eager automaticamente.
    """
    if not enabled or not hasattr(torch, "compile"):
        return model

    # Se estiver no MPS + backend padrão = inductor → não arrisque
    if any(p.device.type == "mps" for p in model.parameters()):
        warnings.warn("torch.compile desabilitado em MPS – usando eager.")
        return model

    try:
        
        compiled = torch.compile(model, mode=mode, backend=backend)
        if hasattr(compiled, "train"):
            return compiled
        warnings.warn("torch.compile devolveu callable sem .train – fallback eager.")
    except Exception as exc:
        warnings.warn(f"torch.compile falhou ({exc}); fallback eager.")

    return model

TEACHER_SIZE = 518

def add_depth_noise(depth, σ_pix=0.12, p_salt=1e-3, blur_ksize=3):
    """depth (B,1,H,W) → depth_noisy (same shape)"""
    B, _, H, W = depth.shape
    depth_np = depth.cpu().numpy()
    out = []
    for d in depth_np:
        d = cv2.GaussianBlur(d[0], (blur_ksize, blur_ksize), 0)
        out.append(d)
    depth_blur = torch.from_numpy(np.stack(out, 0)).unsqueeze(1).to(depth.device)

    g = torch.randn_like(depth_blur) * (depth_blur.abs() * σ_pix + 1e-3)
    noisy = depth_blur + g
    if p_salt:
        mask = torch.rand_like(noisy) < p_salt
        noisy = torch.where(mask, torch.rand_like(noisy) * depth_blur.max(), noisy)
    return noisy


import math, torch
import torch.nn as nn
import torch.nn.functional as F

# ───────────────────────── helpers ───────────────────────────────────────
class ConvGNReLU(nn.Sequential):
    def __init__(self, in_ch, out_ch, k=3, s=1, p=1, groups=8):
        super().__init__(
            nn.Conv2d(in_ch, out_ch, k, s, p, bias=False),
            nn.GroupNorm(groups, out_ch),
            nn.ReLU(inplace=True),
        )

class DWConvGNReLU(nn.Sequential):
    def __init__(self, in_ch, out_ch, k=3, s=1, p=1):
        super().__init__(
            nn.Conv2d(in_ch, in_ch, k, s, p, groups=in_ch, bias=False),
            nn.GroupNorm(max(1, in_ch // 4), in_ch),
            nn.ReLU(inplace=True),
            nn.Conv2d(in_ch, out_ch, 1, bias=False),
            nn.GroupNorm(max(1, out_ch // 4), out_ch),
            nn.ReLU(inplace=True),
        )

# ───────────────────────── ARCH #1 : ROBUST ──────────────────────────────
class _UpBlock(nn.Module):
    def __init__(self, in_ch, skip_ch, out_ch, p_drop):
        super().__init__()
        self.up   = nn.ConvTranspose2d(in_ch, out_ch, 4, 2, 1)
        self.drop = nn.Dropout2d(p=p_drop)
        self.conv = nn.Sequential(
            ConvGNReLU(out_ch + skip_ch, out_ch),
            ConvGNReLU(out_ch, out_ch),
        )
    def forward(self, x, skip):
        x = self.drop(self.up(x))
        if x.shape[-2:] != skip.shape[-2:]:
            x = F.interpolate(x, skip.shape[-2:], mode="bilinear", align_corners=False)
        return self.conv(torch.cat([x, skip], 1))
# refiner_archs.py  (or wherever the classes live)
import math, torch
import torch.nn as nn
import torch.nn.functional as F
# … helpers stay unchanged …


# --------------------------------------------------------------------------
# shared helper – generic Kaiming initialiser
# --------------------------------------------------------------------------
def _kaiming_all(module: nn.Module):
    for m in module.modules():
        if isinstance(m, nn.Conv2d):
            nn.init.kaiming_uniform_(m.weight, a=math.sqrt(5))
            if m.bias is not None:
                nn.init.zeros_(m.bias)


# refiner_archs.py  ─────────────────────────────────────────────────────────
import math, torch, torch.nn as nn, torch.nn.functional as F
# helpers stay unchanged …  ConvGNReLU, DWConvGNReLU, _UpBlock, _kaiming_all


# ═══════════ 1) ROBUST  ════════════════════════════════════════════════════
class RobustDepthRefiner(nn.Module):
    """Heavy U-Net (≈1.2 M params). 3 heads: Δz, logσ², mask-logits."""
    def __init__(self, in_ch=3, base_ch=32, p_drop=0.2, **kw):
        super().__init__()
        c = base_ch
        # ─ encoder / decoder
        self.e1 = nn.Sequential(ConvGNReLU(in_ch, c),  ConvGNReLU(c, c))
        self.e2 = nn.Sequential(ConvGNReLU(c, c*2, s=2), ConvGNReLU(c*2, c*2))
        self.e3 = nn.Sequential(ConvGNReLU(c*2, c*4, s=2), ConvGNReLU(c*4, c*4))
        self.e4 = nn.Sequential(ConvGNReLU(c*4, c*8, s=2), ConvGNReLU(c*8, c*8))
        self.drop = nn.Dropout2d(p_drop)
        self.d3 = _UpBlock(c*8, c*4, c*4, p_drop)
        self.d2 = _UpBlock(c*4, c*2, c*2, p_drop)
        self.d1 = _UpBlock(c*2, c,   c,   p_drop)
        # ─ heads
        self.head_res   = nn.Conv2d(c, 1, 3, padding=1)   # residual Δz
        self.head_logv  = nn.Conv2d(c, 1, 3, padding=1)   # ½ logσ²
        self.head_mask  = nn.Conv2d(c, 1, 3, padding=1)   # mask **logits**
        self._init()

    def _init(self):
        _kaiming_all(self)
        # specialised init
        nn.init.normal_(self.head_res.weight, 0, 1e-4)
        nn.init.zeros_ (self.head_res.bias)
        nn.init.normal_(self.head_logv.weight, 0, 1e-4)
        nn.init.constant_(self.head_logv.bias, -3)         # σ≈e⁻³
        nn.init.zeros_(self.head_mask.weight)
        nn.init.constant_(self.head_mask.bias, 0.0)        #  p≈0.5 start

    def forward(self, x):
        s1 = self.e1(x); s2 = self.e2(s1); s3 = self.e3(s2)
        b  = self.drop(self.e4(s3))
        x  = self.d3(b, s3); x = self.d2(x, s2); x = self.d1(x, s1)
        res   = self.head_res(x)
        logv  = (2 * self.head_logv(x)).clamp_(-8, 4)      # log σ²
        m_logits = self.head_mask(x)                       # raw logits
        return res, logv, m_logits
# ───────────────────────────────────────────────────────────────────────────


# ═══════════ 2) LITE  ═════════════════════════════════════════════════════
class LiteDepthRefiner(nn.Module):
    """Depth-wise U-Net-lite (≈140 k)."""
    def __init__(self, in_ch=3, mid_ch=12, p_drop=0.1, **kw):
        super().__init__()
        c = mid_ch
        self.enc1 = DWConvGNReLU(in_ch, c)
        self.enc2 = DWConvGNReLU(c, c*2, s=2)
        self.bottleneck = nn.Sequential(nn.Dropout2d(p_drop),
                                        DWConvGNReLU(c*2, c*2))
        self.dec1 = DWConvGNReLU(c*2 + c, c)
        self.head = nn.Conv2d(c, 3, 3, 1, 1)       # [Δz, logσ², mask]
        self._init()

    def _init(self):
        _kaiming_all(self)
        nn.init.normal_(self.head.weight, 0, 1e-4)
        with torch.no_grad():
            self.head.bias[:] = torch.tensor([0.0, -3.0, .4])

    def forward(self, x):
        s1 = self.enc1(x)
        b  = self.bottleneck(self.enc2(s1))
        u1 = F.interpolate(b, scale_factor=2, mode="bilinear", align_corners=False)
        u1 = self.dec1(torch.cat([u1, s1], 1))
        res, raw_logv, m_logits = torch.split(self.head(u1), 1, dim=1)
        logv = F.softplus(raw_logv).clamp(-6, 5)
        return res, logv, m_logits
# ───────────────────────────────────────────────────────────────────────────


# ═══════════ 3) TINY  ═════════════════════════════════════════════════════
class TinyDepthRefiner(nn.Module):
    """Ultra-compact refiner (~10 k)."""
    def __init__(self, in_ch=3, base_ch=8, **kw):
        super().__init__()
        self.reduce = nn.Conv2d(in_ch, base_ch, 1, bias=False)
        self.sep1   = DWConvGNReLU(base_ch, base_ch)
        self.sep2   = DWConvGNReLU(base_ch, base_ch)
        self.head   = nn.Conv2d(base_ch, 3, 1)     # [Δz, logσ², mask]
        self._init()

    def _init(self):
        _kaiming_all(self)
        nn.init.normal_(self.head.weight, 0, 1e-4)
        with torch.no_grad():
            self.head.bias[:] = torch.tensor([0.0, -3.0, 0.0])

    def forward(self, x):
        x = self.sep2(self.sep1(self.reduce(x)))
        res, raw_logv, m_logits = torch.split(self.head(x), 1, dim=1)
        logv = F.softplus(raw_logv).clamp(-6, 5)
        return res, logv, m_logits
# ───────────────────────────────────────────────────────────────────────────

# ─────────────────────────── factory ─────────────────────────────────────
_ARCHS = {
    "robust": RobustDepthRefiner,
    "lite":   LiteDepthRefiner,
    "tiny":   TinyDepthRefiner,
}

def build_refiner(name: str, **kwargs) -> nn.Module:
    """
    Example YAML:
    refiner:
      arch: robust        # robust | lite | tiny
      in_ch: 3
      base_ch: 32         # only for 'robust'
      p_drop: 0.2         # optional
    """
    name = str(name).lower()
    if name not in _ARCHS:
        raise ValueError(f"Unknown refiner '{name}'. "
                         f"Choose one of {list(_ARCHS.keys())}.")
    return _ARCHS[name](**kwargs)


import types, torch.nn as nn

def attach_extra_heads(model, cfg):
    if not cfg.get('model', {}).get('attach_heads', False):
        return

    scratch = model.depth_head.scratch

    # 1) find the high-res conv in output_conv2
    highres_conv = scratch.output_conv2[0]  # the 128→32 conv
    features = {}
    handle = highres_conv.register_forward_hook(
        lambda m, inp, out: features.setdefault('feat_hr', out)
    )
    model._extra_head_handle = handle

    # 2) attach your new heads on that 32-channel high-res feature
    feat_ch = highres_conv.out_channels  # 32
    scratch.head_logstd  = nn.Conv2d(feat_ch, 1, 3, padding=1, bias=True)
    scratch.head_masklog = nn.Conv2d(feat_ch, 1, 3, padding=1, bias=True)
    nn.init.constant_(scratch.head_logstd.bias, -3.0)
    nn.init.zeros_(scratch.head_masklog.bias)

    # 3) wrap forward to always return 4-tuple
    orig_fwd = model.forward
    def forward_with_heads(self, *args, **kwargs):
        features.clear()
        out = orig_fwd(*args, **kwargs)
        depth = out[0] if isinstance(out, tuple) else out
        feat_hr = features.get('feat_hr', None)
        logstd  = scratch.head_logstd(feat_hr)  if feat_hr is not None else None
        masklog = scratch.head_masklog(feat_hr) if feat_hr is not None else None
        teacher_used = depth.clone()
        return depth, logstd, masklog, teacher_used

    model.forward = types.MethodType(forward_with_heads, model)


def attach_refiner(model: nn.Module, cfg: dict) -> None:
    """
    Step 2: if cfg['model']['attach_refiner']=True, project features + upsampled
    (depth, logstd, masklog) into a small 'refiner', and wrap forward to ALWAYS
    return a 4-tuple: (depth_refined, logv_refined, mask_refined, teacher_used=depth_u).
    """
    if not cfg.get('model', {}).get('attach_refiner', False):
        return

    # ensure extra heads exist
    if not cfg.get('model', {}).get('attach_heads', False):
        attach_extra_heads(model, cfg)

    device = next(model.parameters()).device
    mh     = cfg['model']
    scratch = model.depth_head.scratch

    # 1) connector
    conn_ch = mh['connector_ch']
    feat_ch = scratch.output_conv1.in_channels
    scratch.connector = nn.Conv2d(feat_ch, conn_ch, 1, bias=False)

    # 2) build and attach refiner
    ref_cfg = mh['refiner'].copy()
    arch    = ref_cfg.pop('arch')
    in_ch   = 3 + conn_ch
    scratch.refiner = build_refiner(arch, in_ch=in_ch, **ref_cfg).to(device)

    # 3) freeze backbone; unfreeze heads, connector, refiner
    for p in model.parameters():
        p.requires_grad = False
    for name, sub in scratch.named_children():
        if name in ('head_res', 'head_logstd', 'head_masklog', 'connector', 'refiner'):
            for p in sub.parameters():
                p.requires_grad = True

    # 4) wrap forward
    orig_fwd = model.forward
    def forward_with_refiner(self, rgb, *args, **kwargs):
        # A) coarse outputs: depth, logstd, masklog, feat
        depth, logstd, masklog, feat = orig_fwd(rgb, *args, **kwargs)
        B,_,h,w = depth.shape
        H,W    = rgb.shape[-2:]

        # B) upsample to full resolution
        depth_u  = F.interpolate(depth,  (H,W), mode='bilinear', align_corners=False)
        logstd_u = F.interpolate(logstd, (H,W), mode='bilinear', align_corners=False) if logstd is not None else torch.zeros_like(depth_u)
        mask_u   = F.interpolate(masklog,(H,W), mode='bilinear', align_corners=False) if masklog is not None else torch.zeros_like(depth_u)
        feat_u   = F.interpolate(feat,   (H,W), mode='bilinear', align_corners=False) if feat is not None else None

        # C) project features
        proj = scratch.connector(feat_u) if feat_u is not None else torch.zeros(B, conn_ch, H, W, device=device)

        # D) refine
        inp = torch.cat([depth_u, logstd_u, mask_u, proj], dim=1)
        res, logv_r, mask_r = scratch.refiner(inp)
        depth_r = depth_u + res

        # always return 4-tuple: (refined, logv, mask_r, teacher_used=depth_u)
        return depth_r, logv_r, mask_r, depth_u

    model.forward = types.MethodType(forward_with_refiner, model)
# import types
# import torch.nn.functional as F


# def attach_refiner(
#     model: torch.nn.Module,
#     cfg=None
# ):
#     device = next(model.parameters()).device

#     ref_cfg = cfg["model"]["refiner"]
#     refiner = build_refiner(ref_cfg.pop("arch"), **ref_cfg).to(device)
#     model.depth_head.add_module("refiner", refiner)
#     orig_fwd = model.forward

#     # freeze everything but the refiner
#     for p in model.parameters():
#         p.requires_grad = False
#     for p in refiner.parameters():
#         p.requires_grad = True

#     def forward_with_refine(self, rgb, depth_gt=None, teacher_pred=None):
#         """
#         • warm-up: use GT depth only (no teacher)
#         • normal: use cached teacher_pred if passed, otherwise call orig_fwd once
#         """
#         B, _, H, W = rgb.shape

#         # ─── warm-up branch ────────────────────────────────────────────────
#         if depth_gt is not None:
#             noisy = add_depth_noise(depth_gt)  # (B,1,H,W)
#             gray  = rgb.mean(1, keepdim=True)  # (B,1,H,W)
#             flag  = torch.zeros_like(gray)     # (B,1,H,W)
#             # align to GT shape just in case
#             _,_,h0,w0 = noisy.shape
#             gray = F.interpolate(gray, (h0,w0), mode="bilinear", align_corners=False)
#             flag = F.interpolate(flag, (h0,w0), mode="nearest")
#             residual, log_var = self.depth_head.refiner(torch.cat([noisy, gray, flag], dim=1))
#             if residual.shape[2:] != (H, W):
#                 residual = F.interpolate(residual, (H, W), mode="bilinear", align_corners=False)
#                 log_var   = F.interpolate(log_var,   (H, W), mode="bilinear", align_corners=False)
#             return noisy + residual, log_var, None

#         # ─── normal branch ─────────────────────────────────────────────────
#         # 1) either use cached teacher_pred or run coarse model once
#         if teacher_pred is not None:
#             # print('Skipping coarse inference, teacher_pred was given')
#             # teacher_pred: (B,1,h_t,w_t) coming in
#             z_noisy = teacher_pred
#         else:
#             print('Running inference...,, teacher_pred wasnt given')
#             rgb_coarse = F.interpolate(rgb, (TEACHER_SIZE, TEACHER_SIZE),
#                                     mode="bilinear", align_corners=False)
#             with torch.no_grad():
#                 z_out = orig_fwd(rgb_coarse)            # (B,1,ht,wt) or (B,ht,wt)
#             if z_out.ndim == 3:
#                 z_noisy = z_out.unsqueeze(1)
#             else:
#                 z_noisy = z_out
#         # 2) upsample teacher to match rgb
#         z_noisy_up = F.interpolate(z_noisy, (H, W),
#                                 mode="bilinear", align_corners=False)

#         # 3) prepare refiner inputs
#         gray = rgb.mean(1, keepdim=True)               # (B,1,H,W)
#         ids  = torch.arange(B, device=rgb.device) % 6  
#         flag = torch.where(ids==4, 1.0, torch.where(ids==5, -1.0, 0.0))
#         flag_map = flag.view(B,1,1,1).expand(-1,1,H,W)

#         # 4) align spatial dims in case right now gray/flag != z_noisy_up
#         _,_,h_up,w_up = z_noisy_up.shape
#         gray     = F.interpolate(gray,     (h_up,w_up), mode="bilinear", align_corners=False)
#         flag_map = F.interpolate(flag_map, (h_up,w_up), mode="nearest")

#         # 5) run refiner
#         inp = torch.cat([z_noisy_up, gray, flag_map], dim=1)
#         residual, log_var, mask_pred = self.depth_head.refiner(inp)
#         if residual.shape[2:] != (H, W):
#             residual = F.interpolate(residual, (H, W), mode="bilinear", align_corners=False)
#             log_var   = F.interpolate(log_var,   (H, W), mode="bilinear", align_corners=False)
#             mask_pred   = F.interpolate(mask_pred,   (H, W), mode="bilinear", align_corners=False)

#         return z_noisy_up + residual, log_var, z_noisy_up, mask_pred 

#     model.forward = types.MethodType(forward_with_refine, model)



# ──────────────────────────────────────────────────────────────────────
# Re-write of build_model_optim_sched
# ──────────────────────────────────────────────────────────────────────
# ─── helpers.py (or train_utils.py) ──────────────────────────────────────
from panorai.depth import ModelRegistry
import torch

def build_teacher_model(cfg: dict) -> torch.nn.Module:
    """
    Loads DepthAnything-v2 (or whatever `model_name` points to),
    moves to eval-only mode, freezes grads, and returns it.
    """
    teacher = ModelRegistry.load(
        cfg["model_name"],
        dataset = cfg["pretrained_on_dataset"],
        encoder = cfg["encoder"],
        return_model = True,
    ).eval()

        # checkpoint (antes da compilação!)
    if cfg.get("load_from"):
        ckpt = torch.load(cfg["load_from"], map_location='cpu', weights_only=False)
        target = teacher._orig_mod if hasattr(teacher, "_orig_mod") else teacher
        target.load_state_dict(ckpt["model"], strict=False)  # strict=False ignora chaves extras
        print(f"✅ checkpoint {cfg['load_from']} carregado")

    for p in teacher.parameters():
        p.requires_grad = False

    return teacher

# def build_model_optim_sched(cfg: Dict, trainloader: DataLoader):
#     device = cfg.get("device", "mps")
#     torch.set_float32_matmul_precision("high")

#     # ── load teacher (unchanged) ───────────────────────────────────────────
#     model = ModelRegistry.load(
#         cfg["model_name"],
#         max_depth=cfg.get("max_depth", 20.0),
#         dataset=cfg["pretrained_on_dataset"],
#         encoder=cfg["encoder"],
#         return_model=True,
#     ).to(torch.float32).to(device)

#     # ── optionally attach residual refiner ─────────────────────────────────
#     if cfg.get("refine", False):
#         attach_refiner(model, cfg=cfg)
#         # sanity: at least one param from refiner must be trainable
#         assert any(p.requires_grad for n,p in model.named_parameters()
#                 if n.startswith("depth_head.refiner")), "Refiner ended up frozen!"
#     elif cfg.get("freeze", True):
#         _freeze_initial(model)

#     # ── optional torch.compile (safe because model object is intact) ───────
#     if cfg.get("compile", True):
#         model = maybe_compile(model, mode="reduce-overhead")

#     # ── optimiser & scheduler (frozen params are ignored automatically) ────
#     base_lr = float(cfg.get("lr", 1e-4))
#     ref_scale = float(cfg.get("lr_refiner_scale", 1.0))

#     name_of = {p: n for n, p in model.named_parameters()}

#     optim = AdamW(
#         _layerwise_lr(model, base_lr, refiner_scale=ref_scale),
#         betas=(0.9, 0.999),
#         weight_decay=float(cfg.get("weight_decay", 1e-3)),
#     )

#     steps_per_epoch = len(trainloader) // cfg.get("grad_accum", 1)
#     sched = ResettableScheduler(optim, cfg, steps_per_epoch)

#     # ─── clear weight-decay for refiner only ──────────────
#     for g in optim.param_groups:
#         if any('depth_head.refiner' in name_of[p] for p in g['params']):
#             g['weight_decay'] = 0.0

#     return model, optim, sched, AdamW
class NoOpScheduler:
    """Bypass scheduler that does nothing but keeps the interface intact."""
    def __init__(self, optim: AdamW, cfg: Dict, steps_per_epoch: int):
        self.optim = optim
        self.cfg = cfg
        self.spe = max(1, steps_per_epoch)
        self._microstep = 0

    def step(self):
        """No-op step."""
        self._microstep += 1
        # No scheduler step is performed.

    def reset(self):
        """No-op reset."""
        self._microstep = 0

def build_model_optim_sched(cfg: Dict, trainloader: DataLoader):
    device = cfg.get("device", "mps")
    torch.set_float32_matmul_precision("high")

    # ── load your base DAV2 (ViT-L) model ────────────────────────────────
    model = ModelRegistry.load(
        cfg["model_name"],
        max_depth=cfg.get("max_depth", 20.0),
        dataset=cfg["pretrained_on_dataset"],
        encoder=cfg["encoder"],
        return_model=True,
    ).to(torch.float32).to(device)

    # ── Stage 1: attach coarse heads? ────────────────────────────────────
    # controlled by cfg["model"]["attach_heads"]
    if cfg.get("model", {}).get("attach_heads", False):
        attach_extra_heads(model, cfg)

    # ── Stage 2: attach high-res refiner? ───────────────────────────────
    # controlled by cfg["model"]["attach_refiner"]
    if cfg.get("model", {}).get("attach_refiner", False):
        attach_refiner(model, cfg)
        # sanity check: refiner must have trainable params
        assert any(
            p.requires_grad for n,p in model.named_parameters()
            if "refiner" in n
        ), "Refiner ended up frozen!"
    # ── otherwise: freeze everything except any heads you did attach ─────
    else:
        _freeze_initial(model)

    # ── optional torch.compile (no problem since we've only patched on top) ─
    if cfg.get("compile", True):
        model = maybe_compile(model, mode="reduce-overhead")

    # ── optimizer & scheduler setup ──────────────────────────────────────
    base_lr   = float(cfg.get("lr", 1e-4))
    ref_scale = float(cfg.get("lr_refiner_scale", 1.0))
    name_of   = {p: n for n, p in model.named_parameters()}

    optim = AdamW(
        _layerwise_lr(model, base_lr, refiner_scale=ref_scale),
        betas=(0.9, 0.999),
        weight_decay=float(cfg.get("weight_decay", 1e-3)),
    )

    steps_per_epoch = len(trainloader) // cfg.get("grad_accum", 1)
    sched = ResettableScheduler(optim, cfg, steps_per_epoch, resetable=True)
    # sched = NoOpScheduler(optim, cfg, steps_per_epoch)

    

    # ── make sure any refiner group has zero weight-decay ───────────────
    for g in optim.param_groups:
        if any("refiner" in name_of[p] for p in g["params"]):
            g["weight_decay"] = 0.0

    return model, optim, sched, AdamW

# # ════════════════════════════════════════════════════════════════════════════
# # Loss & utilities
# # ════════════════════════════════════════════════════════════════════════════
# def build_loss_fn(cfg: Dict) -> FixedDepthLoss:
#     lc = cfg.get("loss", {})

#     return FixedDepthLoss(
#         # ------------ scalar weights -----------------------------------
#         silog_weight   = lc.get("silog_weight",   1.0),
#         l1_weight      = lc.get("l1_weight",      1.0),
#         grad_weight    = lc.get("grad_weight",    1.0),   # <─ was normal_weight
#         normal_weight  = lc.get("normal_weight",  1.0),
#         smooth_weight  = lc.get("smooth_weight",  1.0),
#         thin_weight    = lc.get("thin_weight",    0.0),
#         nll_weight     = lc.get("nll_weight",    0.1),

#         # ------------ loss toggles -------------------------------------
#         use_l1_loss        = lc.get("use_l1_loss",        True),
#         use_silog_loss     = lc.get("use_silog_loss",     False),
#         use_normal_loss    = lc.get("use_normal_loss",    True),
#         use_gradient       = lc.get("use_gradient",       False),
#         use_smoothness     = lc.get("use_smoothness",     False),
#         use_thin_structure = lc.get("use_thin_structure", False),
#         use_l1_attention   = lc.get("use_l1_attention",   False),
#         use_disparity      = lc.get("use_disparity",      False),

#         # ------------ NEW features -------------------------------------
#         hetero          = lc.get("hetero",          False),  # log-var / NLL
#         silog_lambda    = lc.get("silog_lambda",    0.5),    # keep existing
#         residual_reg_weight = lc.get("residual_reg_weight",      .02),
#         residual_gate_thresh = lc.get("rresidual_gate_thresh",      .2),
#     )

# ════════════════════════════════════════════════════════════════════════════
# Loss & utilities
# ════════════════════════════════════════════════════════════════════════════
def build_loss_fn(cfg: Dict) -> FixedDepthLoss:
    lc = cfg.get("loss", {})

    # helper → zero out if the flag is False, otherwise take the weight
    def _w(key_weight, key_flag, default):
        return 0.0 if not lc.get(key_flag, True) else lc.get(key_weight, default)
    
    loss_fn = FixedDepthLoss(
        # ─── primary weights (flags control zeros) ───
        silog_weight        = _w("silog_weight",  "use_silog_loss",     1.0),
        grad_weight         = _w("grad_weight",   "use_gradient",       0.1),
        smooth_weight       = _w("smooth_weight", "use_smoothness",     0.1),
        thin_weight         = _w("thin_weight",   "use_thin_structure", 0.0),
        normal_weight       = _w("normal_weight", "use_normal_loss",    0.0),
        l1_weight           = _w("l1_weight",     "use_l1_loss",        0.0),

        # ─── L1 extras ───────────────────────────────
        l1_3d_mode          = lc.get("l1_3d_mode",      "none"),
        use_smooth_l1       = lc.get("use_smooth_l1",   True),

        # ─── SiLog & char-b parameters ───────────────
        silog_lambda        = lc.get("silog_lambda",    0.5),

        # ─── gradient / thin hyperparams ────────────
        k_edge_grad         = lc.get("k_edge_grad",     3.0),
        thin_pct            = lc.get("thin_pct",        0.70),

        # ─── hetero & NLL ────────────────────────────
        hetero               = lc.get("hetero",         False),
        nll_weight           = lc.get("nll_weight",     1.0),
        nll_ramp_epochs      = lc.get("nll_ramp_epochs",5),

        # ─── residual / refiner reg ─────────────────
        residual_reg_weight  = lc.get("residual_reg_weight",0.02),
        residual_tau_mode    = lc.get("residual_tau_mode",  "adaptive"),

        # ─── mask‐BCE head ───────────────────────────
        use_mask_bce         = lc.get("use_mask_bce",    False),
        mask_weight          = lc.get("mask_weight",     0.2),

        # ─── 3D‐L1 weighting ─────────────────────────
        sixth_weight         = lc.get("sixth_weight",    0.2),

        # ─── scheduling (warm‐up/ramp) ───────────────
        schedule_keys        = lc.get("schedule_keys",   ["thin_w", "normal_w"]),
        warmup_epochs        = lc.get("warmup_epochs",   3),
        ramp_epochs          = lc.get("ramp_epochs",     0),
    )
    
    try:
        return torch.jit.script(loss_fn)
    except Exception as e:
        print(['[jit compile failed]: {e}'])
        return loss_fn

# def build_loss_fn(cfg: Dict) -> FixedDepthLoss:
#     lc = cfg.get("loss", {})

#     # helper → weight = 0 if component switched off
#     def _w(key_weight, key_flag, default):
#         return 0.0 if not lc.get(key_flag, True) else lc.get(key_weight, default)

#     return FixedDepthLoss(
#         # ─── primary weights (flags control zeros) ───
#         silog_weight  = _w("silog_weight",  "use_silog_loss",  1.0),
#         l1_weight     = _w("l1_weight",     "use_l1_loss",     1.0),
#         grad_weight   = _w("grad_weight",   "use_gradient",    0.1),
#         smooth_weight = _w("smooth_weight", "use_smoothness",  0.1),
#         thin_weight   = _w("thin_weight",   "use_thin_structure", 0.0),
#         normal_weight = _w("normal_weight", "use_normal_loss", 1.0),

#         # ─── heteroscedastic & residual stuff ───
#         hetero                = lc.get("hetero", False),
#         nll_weight            = lc.get("nll_weight", 0.5),
#         nll_ramp_epochs       = lc.get("nll_ramp_epochs", 5),
#         residual_reg_weight   = lc.get("residual_reg_weight", 0.0),
#         residual_tau_mode     = "fixed",   # kept simple; change if you need adaptive
#         k_edge_grad           = 3.0,
#         thin_pct              = 0.70,

#         # ─── L1 extras ───
#         l1_3d_mode      = lc.get("l1_3d_mode", "log"),
#         sixth_weight    = lc.get("sixth_weight", 0.2),
#         use_smooth_l1   = lc.get("use_smooth_l1", True),

#         # ─── mask term ───
#         use_mask_bce = lc.get("use_mask_bce", False),
#         mask_weight  = lc.get("mask_weight", 0.2),
#     )
def search_best_scale(model, trainloader: DataLoader, cfg: Dict) -> float:
    return estimate_best_scale(
        model, trainloader,
        device=cfg.get("device", "mps"),
        max_depth=cfg.get("max_depth", 80.0),
        robust=True, plot=True,
    )

# ════════════════════════════════════════════════════════════════════════════
# TransformerUnfreezeScheduler
# ════════════════════════════════════════════════════════════════════════════
# class TransformerUnfreezeScheduler:
#     """Validation/gradient-driven unfreeze with boom-log."""

#     def __init__(
#         self,
#         model: torch.nn.Module,
#         *,
#         start_block: int = 23,
#         min_block: int = 17,
#         grad_threshold: float = 1e-3,
#         patience: int = 10,
#         warmup_epochs: int = 5,
#         min_epochs_per_unfreeze: int = 5,
#         log_path: str | Path = "unfreeze_log.jsonl",
#         logger=print,
#     ) -> None:
#         self.m = model
#         self.sblk = start_block
#         self.cur_min = start_block
#         self.min_blk = min_block
#         self.grad_thr = grad_threshold
#         self.patience = patience
#         self.warmup = warmup_epochs
#         self.min_epochs = min_epochs_per_unfreeze
#         self.best_val = float("inf")
#         self.wait = 0
#         self.epoch = 0
#         self.last_unfreeze_epoch = 0
#         self.log = logger
#         self.log_path = Path(log_path)
#         self.log_path.parent.mkdir(parents=True, exist_ok=True)

#     # ——— trainer hooks ———
#     def update(self, *, epoch: int, model, val_logs: Dict, **_):
#         self.epoch = epoch
#         self._step_validation(val_logs.get("total", float("inf")))
#         self._step_gradients(model)

#     def step_gradients(self, model):
#         self._step_gradients(model)

#     # ——— internal ———
#     def _can_unfreeze(self):
#         return (self.epoch - self.last_unfreeze_epoch) >= self.min_epochs and self.cur_min > self.min_blk

#     def _step_validation(self, val_loss: float):
#         if self.epoch < self.warmup or not np.isfinite(val_loss):
#             return
#         if val_loss < self.best_val - 1e-4:
#             self.best_val, self.wait = val_loss, 0
#         else:
#             self.wait += 1
#         if self.wait >= self.patience and self._can_unfreeze():
#             self.cur_min -= 1
#             self.wait = 0
#             self._unfreeze_range(self.cur_min, self.sblk, "val_plateau")

#     def _step_gradients(self, model):
#         if not self._can_unfreeze():
#             return
#         blk = f"pretrained.blocks.{self.cur_min}"
#         grads = [p.grad.norm().item() for n, p in model.named_parameters()
#                  if n.startswith(blk) and p.grad is not None]
#         if grads and max(grads) < self.grad_thr:
#             self.cur_min -= 1
#             self._unfreeze_range(self.cur_min, self.sblk, "weak_grad")

#     def _unfreeze_range(self, mn: int, mx: int, reason: str):
#         if hasattr(self, "_rebuild"):
#             self._rebuild()  
# #        also reset the LR schedule so unfrozen blocks get a fresh cycle
#         try:
#             self._rebuild_trainer.scheduler.reset()
#         except Exception:
#             pass

#         names: List[str] = []
#         for n, p in self.m.named_parameters():
#             if any(n.startswith(f"pretrained.blocks.{i}") for i in range(mn, mx + 1)) and not p.requires_grad:
#                 p.requires_grad = True
#                 names.append(n)
#         if not names:
#             return
#         self.last_unfreeze_epoch = self.epoch
#         self._boom_log(names, reason, mn, mx)

#     def _boom_log(self, names: List[str], reason: str, mn: int, mx: int):
#         msg = (f"🔓 epoch={self.epoch} | blocks {mn}–{mx} | "
#                f"layers={len(names)} | reason={reason}")
#         self.log(msg)
#         evt = {
#             "epoch": self.epoch,
#             "reason": reason,
#             "blocks": [mn, mx],
#             "layers": names,
#             "time": time.strftime("%Y-%m-%d %H:%M:%S"),
#         }
#         with self.log_path.open("a") as fp:
#             fp.write(json.dumps(evt) + "\n")

#     def set_logger(self, logger_fn):
#         self.log = logger_fn

#     def set_rebuild_callback(self, fn):
#         self._rebuild = fn

class TransformerUnfreezeScheduler:
    """Validation/gradient-driven unfreeze with boom-log, one group per epoch."""

    def __init__(
        self,
        model: torch.nn.Module,
        *,
        start_block: int = 23,
        min_block: int = 17,
        grad_threshold: float = 1e-3,
        patience: int = 10,
        warmup_epochs: int = 5,
        min_epochs_per_unfreeze: int = 5,
        log_path: str | Path = "unfreeze_log.jsonl",
        logger=print,
    ) -> None:
        self.m = model
        self.grad_thr = grad_threshold
        self.patience = patience
        self.warmup = warmup_epochs
        self.min_epochs = min_epochs_per_unfreeze
        self.best_val = float("inf")
        self.wait = 0
        self.epoch = 0
        self.last_unfreeze_epoch = -1  # allow unfreezing on epoch 0 if ready
        self.log = logger
        self.log_path = Path(log_path)
        self.log_path.parent.mkdir(parents=True, exist_ok=True)

        # ─── fresh start: delete any existing log from previous runs ───
        if self.log_path.exists():
            self.log_path.unlink()

        # Define unfreeze groups: norm, head_conv2, head_conv1, then backbone blocks
        self.unfreeze_groups: List[Dict[str, List[str]]] = []
        self.unfreeze_groups.append({"name": "norm",       "prefixes": ["pretrained.norm."]})
        self.unfreeze_groups.append({"name": "head_conv2", "prefixes": ["depth_head.scratch.output_conv2."]})
        self.unfreeze_groups.append({"name": "head_conv1", "prefixes": ["depth_head.scratch.output_conv1"]})
        for b in range(start_block, min_block - 1, -1):
            self.unfreeze_groups.append({
                "name": f"block{b}",
                "prefixes": [f"pretrained.blocks.{b}."]
            })

        self.next_group = 0
        self._rebuild = None

    def update(self, *, epoch: int, model, val_logs: Dict, **_):
        self.epoch = epoch
        self._step_validation(val_logs.get("total", float("inf")))
        self._step_gradients(model)

    def step_gradients(self, model):
        self._step_gradients(model)

    def _can_unfreeze(self) -> bool:
        return (
            self.next_group < len(self.unfreeze_groups)
            and (self.epoch - self.last_unfreeze_epoch) >= self.min_epochs
            and self.epoch >= self.warmup
        )

    def _step_validation(self, val_loss: float) -> None:
        if self.epoch < self.warmup or not np.isfinite(val_loss):
            return
        if val_loss < self.best_val - 1e-4:
            self.best_val, self.wait = val_loss, 0
        else:
            self.wait += 1
        if self.wait >= self.patience and self._can_unfreeze():
            if self._unfreeze_next_group("val_plateau"):
                self.wait = 0

    def _step_gradients(self, model) -> None:
        if not self._can_unfreeze():
            return
        grp = self.unfreeze_groups[self.next_group]
        grads = [
            p.grad.norm().item()
            for n, p in model.named_parameters()
            if any(n.startswith(pref) for pref in grp["prefixes"]) and p.grad is not None
        ]
        if grads and max(grads) < self.grad_thr:
            self._unfreeze_next_group("weak_grad")

    def _unfreeze_next_group(self, reason: str) -> bool:
        if self.next_group >= len(self.unfreeze_groups):
            return False

        grp = self.unfreeze_groups[self.next_group]

        # Rebuild optimizer/scheduler if needed
        if self._rebuild:
            self._rebuild()
            try:
                trainer = self._rebuild.__self__
                trainer.scheduler.reset()
            except Exception:
                pass

        names = []
        for n, p in self.m.named_parameters():
            if any(n.startswith(pref) for pref in grp["prefixes"]) and not p.requires_grad:
                p.requires_grad = True
                names.append(n)

        if not names:
            self.next_group += 1
            return False

        if self.last_unfreeze_epoch == self.epoch:
            self.log(f"⚠️ Already unfrozen a group in epoch={self.epoch}, skipping.")
            return False

        self.last_unfreeze_epoch = self.epoch
        self._boom_log(names, reason, grp["name"])
        self.next_group += 1
        return True

    def _boom_log(self, names: List[str], reason: str, group: str) -> None:
        msg = f"🔓 epoch={self.epoch} | group={group} | layers={len(names)} | reason={reason}"
        self.log(msg)
        evt = {
            "epoch": self.epoch,
            "group": group,
            "reason": reason,
            "layers": names,
            "time": time.strftime("%Y-%m-%d %H:%M:%S"),
        }
        with self.log_path.open("a") as fp:
            fp.write(json.dumps(evt) + "\n")

    def set_logger(self, logger_fn) -> None:
        self.log = logger_fn

    def set_rebuild_callback(self, fn) -> None:
        self._rebuild = fn

# EOF




class EMAAdaptiveClipper:
    def __init__(self, decay=0.99, max_factor=2.0, min_scale=0.25):
        self.decay = decay
        self.max_factor = max_factor
        self.min_scale = min_scale
        self.ema_norms = {}  # name → running norm


    def step(self, model: torch.nn.Module, verbose=False):
        clipped = []
        for name, param in model.named_parameters():
            if param.grad is None:
                continue

            gnorm = param.grad.norm().item()
            ema = self.ema_norms.get(name, gnorm)
            ema = self.decay * ema + (1 - self.decay) * gnorm
            self.ema_norms[name] = ema

            if gnorm > self.max_factor * ema:
                ratio = gnorm / ema
                scale = 1 / (1 + (ratio - 1) ** 2)
                scale = max(scale, self.min_scale)
                param.grad.mul_(scale)
                clipped.append((name, gnorm, ema, scale))

        if verbose and clipped:
            print("[EMA Adaptive Clipping Summary]")
            for name, g, ema, s in clipped:
                print(f"  [clip] {name:40s} grad={g:.2f} ema={ema:.2f} scale=×{s:.3f}")





# EOF


# Copyright (C) 2024-present Naver Corporation. All rights reserved.
# Licensed under CC BY-NC-SA 4.0 (non-commercial use only).
#
# --------------------------------------------------------
# post process function for all heads: extract 3D points/confidence from output
# --------------------------------------------------------
import torch


def postprocess(out, depth_mode, conf_mode):
    """
    extract 3D points/confidence from prediction head output
    """
    fmap = out.permute(0, 2, 3, 1)  # B,H,W,3
    res = dict(pts3d=reg_dense_depth(fmap[:, :, :, 0:3], mode=depth_mode))

    if conf_mode is not None:
        res['conf'] = reg_dense_conf(fmap[:, :, :, 3], mode=conf_mode)
    return res


def reg_dense_depth(xyz, mode):
    """
    extract 3D points from prediction head output
    """
    mode, vmin, vmax = mode

    no_bounds = (vmin == -float('inf')) and (vmax == float('inf'))
    assert no_bounds

    if mode == 'linear':
        if no_bounds:
            return xyz  # [-inf, +inf]
        return xyz.clip(min=vmin, max=vmax)

    # distance to origin
    d = xyz.norm(dim=-1, keepdim=True)
    xyz = xyz / d.clip(min=1e-8)

    if mode == 'square':
        return xyz * d.square()

    if mode == 'exp':
        return xyz * torch.expm1(d)

    raise ValueError(f'bad {mode=}')


def reg_dense_conf(x, mode):
    """
    extract confidence from prediction head output
    """
    mode, vmin, vmax = mode
    if mode == 'exp':
        return vmin + x.exp().clip(max=vmax-vmin)
    if mode == 'sigmoid':
        return (vmax - vmin) * torch.sigmoid(x) + vmin
    raise ValueError(f'bad {mode=}')



import torch


def eval_depth(pred, target):
    assert pred.shape == target.shape

    thresh = torch.max((target / pred), (pred / target))

    d1 = torch.sum(thresh < 1.25).float() / len(thresh)
    d2 = torch.sum(thresh < 1.25 ** 2).float() / len(thresh)
    d3 = torch.sum(thresh < 1.25 ** 3).float() / len(thresh)

    diff = pred - target
    diff_log = torch.log(pred) - torch.log(target)

    abs_rel = torch.mean(torch.abs(diff) / target)
    sq_rel = torch.mean(torch.pow(diff, 2) / target)

    rmse = torch.sqrt(torch.mean(torch.pow(diff, 2)))
    rmse_log = torch.sqrt(torch.mean(torch.pow(diff_log , 2)))

    log10 = torch.mean(torch.abs(torch.log10(pred) - torch.log10(target)))
    silog = torch.sqrt(torch.pow(diff_log, 2).mean() - 0.5 * torch.pow(diff_log.mean(), 2))

    return {'d1': d1.item(), 'd2': d2.item(), 'd3': d3.item(), 'abs_rel': abs_rel.item(), 'sq_rel': sq_rel.item(), 
            'rmse': rmse.item(), 'rmse_log': rmse_log.item(), 'log10':log10.item(), 'silog':silog.item()}


import torch
import torch.nn as nn
import numpy as np


class VNLoss(nn.Module):
    """
    Virtual Normal Loss.
    """
    def __init__(self,
                 delta_cos=0.867, delta_diff_x=0.01,
                 delta_diff_y=0.01, delta_diff_z=0.01,
                 delta_z=1e-5, sample_ratio=0.15,
                 loss_weight=1.0, data_type=['sfm', 'stereo', 'lidar', 'denselidar', 'denselidar_nometric', 'denselidar_syn'], **kwargs):
        super(VNLoss, self).__init__()
        self.delta_cos = delta_cos
        self.delta_diff_x = delta_diff_x
        self.delta_diff_y = delta_diff_y
        self.delta_diff_z = delta_diff_z
        self.delta_z = delta_z
        self.sample_ratio = sample_ratio
        self.loss_weight = loss_weight
        self.data_type = data_type
        self.eps = 1e-6


    def init_image_coor(self, intrinsic, height, width):
        # x_row = torch.arange(0, W, device="cuda")
        # x = torch.tile(x_row, (H, 1))
        # x = x.to(torch.float32)
        # u_m_u0 = x[None, None, :, :] - u0
        # self.register_buffer('u_m_u0', u_m_u0, persistent=False)

        # y_col = torch.arange(0, H, device="cuda")  # y_col = np.arange(0, height)
        # y = torch.transpose(torch.tile(y_col, (W, 1)), 1, 0)
        # y = y.to(torch.float32)
        # v_m_v0 = y[None, None, :, :] - v0
        # self.register_buffer('v_m_v0', v_m_v0, persistent=False)

        # pix_idx_mat = torch.arange(H*W, device="cuda").reshape((H, W))
        # self.register_buffer('pix_idx_mat', pix_idx_mat, persistent=False)
        #self.pix_idx_mat = torch.arange(height*width, device="cuda").reshape((height, width))
        
        u0 = intrinsic[:, 0, 2][:, None, None, None]
        v0 = intrinsic[:, 1, 2][:, None, None, None]
        y, x = torch.meshgrid([torch.arange(0, height, dtype=torch.float32, device="cuda"),
                               torch.arange(0, width, dtype=torch.float32, device="cuda")], indexing='ij')
        u_m_u0 = x[None, None, :, :] - u0
        v_m_v0 = y[None, None, :, :] - v0
        # return u_m_u0, v_m_v0
        self.register_buffer('v_m_v0', v_m_v0, persistent=False)
        self.register_buffer('u_m_u0', u_m_u0, persistent=False)

    def transfer_xyz(self, depth, focal_length, u_m_u0, v_m_v0):
        x = u_m_u0 * depth / focal_length
        y = v_m_v0 * depth / focal_length
        z = depth
        pw = torch.cat([x, y, z], 1).permute(0, 2, 3, 1).contiguous() # [b, h, w, c]
        return pw

    def select_index(self, B, H, W, mask):
        """
        
        """
        p1 = []
        p2 = []
        p3 = []
        pix_idx_mat = torch.arange(H*W, device="cuda").reshape((H, W))
        for i in range(B):
            inputs_index = torch.masked_select(pix_idx_mat, mask[i, ...].gt(self.eps))
            num_effect_pixels = len(inputs_index)

            intend_sample_num = int(H * W * self.sample_ratio)
            sample_num = intend_sample_num if num_effect_pixels >= intend_sample_num else num_effect_pixels

            shuffle_effect_pixels = torch.randperm(num_effect_pixels, device="cuda")
            p1i = inputs_index[shuffle_effect_pixels[:sample_num]]
            shuffle_effect_pixels = torch.randperm(num_effect_pixels, device="cuda")
            p2i = inputs_index[shuffle_effect_pixels[:sample_num]]
            shuffle_effect_pixels = torch.randperm(num_effect_pixels, device="cuda")
            p3i = inputs_index[shuffle_effect_pixels[:sample_num]]

            cat_null = torch.tensor(([0,] * (intend_sample_num - sample_num)), dtype=torch.long, device="cuda")
            p1i = torch.cat([p1i, cat_null])
            p2i = torch.cat([p2i, cat_null])
            p3i = torch.cat([p3i, cat_null])

            p1.append(p1i)
            p2.append(p2i)
            p3.append(p3i)
        
        p1 = torch.stack(p1, dim=0)
        p2 = torch.stack(p2, dim=0)
        p3 = torch.stack(p3, dim=0)

        p1_x = p1 % W
        p1_y = torch.div(p1, W, rounding_mode='trunc').long() # p1 // W

        p2_x = p2 % W
        p2_y = torch.div(p2, W, rounding_mode='trunc').long() # p2 // W

        p3_x = p3 % W
        p3_y = torch.div(p3, W, rounding_mode='trunc').long() # p3 // W
        p123 = {'p1_x': p1_x, 'p1_y': p1_y, 'p2_x': p2_x, 'p2_y': p2_y, 'p3_x': p3_x, 'p3_y': p3_y}
        return p123

    def form_pw_groups(self, p123, pw):
        """
        Form 3D points groups, with 3 points in each grouup.
        :param p123: points index
        :param pw: 3D points
        :return:
        """
        B, _, _, _ = pw.shape
        p1_x = p123['p1_x']
        p1_y = p123['p1_y']
        p2_x = p123['p2_x']
        p2_y = p123['p2_y']
        p3_x = p123['p3_x']
        p3_y = p123['p3_y']
        
        pw_groups = []
        for i in range(B):
            pw1 = pw[i, p1_y[i], p1_x[i], :]
            pw2 = pw[i, p2_y[i], p2_x[i], :]
            pw3 = pw[i, p3_y[i], p3_x[i], :]
            pw_bi = torch.stack([pw1, pw2, pw3], dim=2)
            pw_groups.append(pw_bi)
        # [B, N, 3(x,y,z), 3(p1,p2,p3)]
        pw_groups = torch.stack(pw_groups, dim=0)
        return pw_groups

    def filter_mask(self, p123, gt_xyz, delta_cos=0.867,
                    delta_diff_x=0.005,
                    delta_diff_y=0.005,
                    delta_diff_z=0.005):
        pw = self.form_pw_groups(p123, gt_xyz)
        pw12 = pw[:, :, :, 1] - pw[:, :, :, 0]
        pw13 = pw[:, :, :, 2] - pw[:, :, :, 0]
        pw23 = pw[:, :, :, 2] - pw[:, :, :, 1]
        ###ignore linear
        pw_diff = torch.cat([pw12[:, :, :, np.newaxis], pw13[:, :, :, np.newaxis], pw23[:, :, :, np.newaxis]],
                            3)  # [b, n, 3, 3]
        m_batchsize, groups, coords, index = pw_diff.shape
        proj_query = pw_diff.view(m_batchsize * groups, -1, index).permute(0, 2, 1).contiguous()  # (B* X CX(3)) [bn, 3(p123), 3(xyz)]
        proj_key = pw_diff.contiguous().view(m_batchsize * groups, -1, index)  # B X  (3)*C [bn, 3(xyz), 3(p123)]
        q_norm = proj_query.norm(2, dim=2)
        nm = torch.bmm(q_norm.contiguous().view(m_batchsize * groups, index, 1), q_norm.view(m_batchsize * groups, 1, index)) #[]
        energy = torch.bmm(proj_query, proj_key)  # transpose check [bn, 3(p123), 3(p123)]
        norm_energy = energy / (nm + self.eps)
        norm_energy = norm_energy.contiguous().view(m_batchsize * groups, -1)
        mask_cos = torch.sum((norm_energy > delta_cos) + (norm_energy < -delta_cos), 1) > 3  # igonre
        mask_cos = mask_cos.contiguous().view(m_batchsize, groups)
        ##ignore padding and invilid depth
        mask_pad = torch.sum(pw[:, :, 2, :] > self.delta_z, 2) == 3

        ###ignore near
        mask_x = torch.sum(torch.abs(pw_diff[:, :, 0, :]) < delta_diff_x, 2) > 0
        mask_y = torch.sum(torch.abs(pw_diff[:, :, 1, :]) < delta_diff_y, 2) > 0
        mask_z = torch.sum(torch.abs(pw_diff[:, :, 2, :]) < delta_diff_z, 2) > 0

        mask_ignore = (mask_x & mask_y & mask_z) | mask_cos
        mask_near = ~mask_ignore
        mask = mask_pad & mask_near

        return mask, pw

    def select_points_groups(self, gt_depth, pred_depth, intrinsic, mask):
        B, C, H, W = gt_depth.shape
        focal_length = intrinsic[:, 0, 0][:, None, None, None]
        u_m_u0, v_m_v0 = self.u_m_u0, self.v_m_v0 # self.init_image_coor(intrinsic, height=H, width=W)
        
        pw_gt = self.transfer_xyz(gt_depth, focal_length, u_m_u0, v_m_v0)
        pw_pred = self.transfer_xyz(pred_depth, focal_length, u_m_u0, v_m_v0)
        
        p123 = self.select_index(B, H, W, mask)
        # mask:[b, n], pw_groups_gt: [b, n, 3(x,y,z), 3(p1,p2,p3)]
        mask, pw_groups_gt = self.filter_mask(p123, pw_gt,
                                              delta_cos=0.867,
                                              delta_diff_x=0.005,
                                              delta_diff_y=0.005,
                                              delta_diff_z=0.005)

        # [b, n, 3, 3]
        pw_groups_pred = self.form_pw_groups(p123, pw_pred)
        pw_groups_pred[pw_groups_pred[:, :, 2, :] == 0] = 0.0001
        mask_broadcast = mask.repeat(1, 9).reshape(B, 3, 3, -1).permute(0, 3, 1, 2).contiguous()
        pw_groups_pred_not_ignore = pw_groups_pred[mask_broadcast].reshape(1, -1, 3, 3)
        pw_groups_gt_not_ignore = pw_groups_gt[mask_broadcast].reshape(1, -1, 3, 3)

        return pw_groups_gt_not_ignore, pw_groups_pred_not_ignore

    def forward(self, prediction, target, mask, intrinsic, select=True, **kwargs): #gt_depth, pred_depth, select=True):
        """
        Virtual normal loss.
        :param prediction: predicted depth map, [B,W,H,C]
        :param data: target label, ground truth depth, [B, W, H, C], padding region [padding_up, padding_down]
        :return:
        """
        loss  = self.get_loss(prediction, target, mask, intrinsic, select, **kwargs)
        return loss
 
    
    def get_loss(self, prediction, target, mask, intrinsic, select=True, **kwargs):
        # configs for the cameras
        # focal_length = intrinsic[:, 0, 0][:, None, None, None]
        # u0 = intrinsic[:, 0, 2][:, None, None, None]
        # v0 = intrinsic[:, 1, 2][:, None, None, None]
        B, _, H, W = target.shape
        if 'u_m_u0' not in self._buffers or 'v_m_v0' not in self._buffers \
            or self.u_m_u0.shape != torch.Size([B,1,H,W]) or self.v_m_v0.shape != torch.Size([B,1,H,W]):
            self.init_image_coor(intrinsic, H, W)


        gt_points, pred_points = self.select_points_groups(target, prediction, intrinsic, mask)

        gt_p12 = gt_points[:, :, :, 1] - gt_points[:, :, :, 0]
        gt_p13 = gt_points[:, :, :, 2] - gt_points[:, :, :, 0]
        pred_p12 = pred_points[:, :, :, 1] - pred_points[:, :, :, 0]
        pred_p13 = pred_points[:, :, :, 2] - pred_points[:, :, :, 0]

        gt_normal = torch.cross(gt_p12, gt_p13, dim=2)
        pred_normal = torch.cross(pred_p12, pred_p13, dim=2)
        pred_norm = torch.norm(pred_normal, 2, dim=2, keepdim=True)
        gt_norm = torch.norm(gt_normal, 2, dim=2, keepdim=True)
        pred_mask = pred_norm == 0.0
        gt_mask = gt_norm == 0.0
        pred_mask = pred_mask.to(torch.float32)
        gt_mask = gt_mask.to(torch.float32)
        pred_mask *= self.eps
        gt_mask *= self.eps
        gt_norm = gt_norm + gt_mask
        pred_norm = pred_norm + pred_mask
        gt_normal = gt_normal / gt_norm
        pred_normal = pred_normal / pred_norm
        loss = torch.abs(gt_normal - pred_normal)
        loss = torch.sum(torch.sum(loss, dim=2), dim=0)
        if select:
            loss, indices = torch.sort(loss, dim=0, descending=False)
            loss = loss[int(loss.size(0) * 0.25):]
        loss = torch.sum(loss) / (loss.numel() + self.eps)
        if torch.isnan(loss).item() | torch.isinf(loss).item():
            loss = 0 * torch.sum(prediction)
            print(f'VNL NAN error, {loss}')        
        return loss * self.loss_weight


if __name__ == '__main__':
    import cv2
    vnl_loss = VNLoss()
    pred_depth = np.random.random([2, 1, 480, 640])
    gt_depth = np.zeros_like(pred_depth) #np.random.random([2, 1, 480, 640])
    intrinsic = [[[100, 0, 200], [0, 100, 200], [0, 0, 1]], [[100, 0, 200], [0, 100, 200], [0, 0, 1]],]
    gt_depth = torch.tensor(np.array(gt_depth, np.float32)).cuda()
    pred_depth = torch.tensor(np.array(pred_depth, np.float32)).cuda()
    intrinsic = torch.tensor(np.array(intrinsic, np.float32)).cuda()
    mask = gt_depth > 0
    loss1 = vnl_loss(pred_depth, gt_depth, mask, intrinsic)
    loss2 = vnl_loss(pred_depth, gt_depth, mask, intrinsic)
    print(loss1, loss2)



from panorai import EquirectangularImage
import numpy as np
import random
import cv2

import numpy as np
from panorai import EquirectangularImage
import random

import numpy as np
from panorai import EquirectangularImage
import os

class DeterministicProjectionTransform:
    """
    Torch-style transform that deterministically applies:
    - longitude shift (angle_idx)
    - horizontal flip
    - gnomonic projection

    Args:
        size (int): Output face size (square).
        n_angles (int): Number of longitude buckets between 0–45°.
    """
    def __init__(self, size=518, n_angles=8):
        self.size = size
        self.n_angles = n_angles
        self.lon_angles = np.linspace(0, 45, n_angles)

    def __call__(self, sample_with_key):
        # Expect: { 'key': str, 'data': sample_dict }
        key = sample_with_key['key']
        sample = sample_with_key['data']

        idx = int(key.split('_')[0])
        angle_idx = int(key.split('_a')[1].split('_')[0])
        flip = int(key.split('_f')[1]) == 1

        delta_lon = float(self.lon_angles[angle_idx])
        H = W = 14 * (self.size // 14)

        out = {}
        for k, v in sample.items():
            if flip:
                v = np.flip(v, axis=-1)

            e = EquirectangularImage(v)
            e.preprocess(shadow_angle=30, delta_lat=0.0, delta_lon=delta_lon)
            e.attach_projection("gnomonic", x_points=W, y_points=H)
            faceset = np.array(e.to_gnomonic_face_set())

            if 'xyz' in k:
                u, v_ = np.meshgrid(np.linspace(-1, 1, W), np.linspace(-1, 1, H), indexing='xy')
                faceset = np.sqrt(faceset**2 / (u[None]**2 + v_[None]**2 + 1))

            out[k] = np.ascontiguousarray(faceset.astype(np.float32))

        return out
    
class DeterministicAugmentor:
    def __init__(self, size, n_angles=8, enable_flip=True):
        self.size = size
        self.n_angles = n_angles
        self.lon_angles = np.linspace(0, 45, n_angles)
        self.enable_flip = enable_flip

    def __call__(self, sample, angle_idx=0, flip_idx=0):
        """
        Args:
            sample (dict): {'rgb_image': ..., 'xyz_image': ...}
            angle_idx (int): which angle bucket to use (0 to n_angles-1)
            flip_idx (int): 0 = no flip, 1 = flip
        """
        delta_lon = float(self.lon_angles[angle_idx])
        do_flip = (flip_idx == 1 and self.enable_flip)

        new_sample = {}
        H, W = 14 * (self.size // 14), 14 * (self.size // 14)

        for k, v in sample.items():
            if do_flip:
                v = np.flip(v, axis=-1)

            e = EquirectangularImage(v)
            e.preprocess(shadow_angle=30, delta_lat=0.0, delta_lon=delta_lon)
            e.attach_projection("gnomonic", x_points=W, y_points=H)
            faceset = np.array(e.to_gnomonic_face_set())

            if 'xyz' in k:
                u, v_ = np.meshgrid(
                    np.linspace(-1, 1, W),
                    np.linspace(-1, 1, H),
                    indexing='xy'
                )
                faceset = np.sqrt(faceset**2 / (u[None]**2 + v_[None]**2 + 1))

            new_sample[k] = np.array(faceset).astype(np.float32)
            new_sample[k] = np.ascontiguousarray(new_sample[k])

        return new_sample
    
import numpy as np
import random
from panorai import EquirectangularImage


import torch
import torch.nn.functional as F

import numpy as np
import cv2

def zero_depth_edges_batch(depth_maps, ksize=3, threshold=10):
    """
    Applies Sobel edge detection on a batch of depth maps and zeroes out edges.

    Args:
        depth_maps (np.ndarray): Batch of depth maps, shape (B, H, W).
        ksize (int): Sobel kernel size (must be 1, 3, 5, or 7).
        threshold (float): Gradient magnitude threshold for edge detection.

    Returns:
        np.ndarray: Depth maps with edges zeroed out (same shape as input).
    """
    batch_out = []

    for depth in depth_maps:
        # Convert to float32 for precision
        depth_f = depth.astype(np.float32)

        # Compute Sobel gradients
        grad_x = cv2.Sobel(depth_f, cv2.CV_32F, 1, 0, ksize=ksize)
        grad_y = cv2.Sobel(depth_f, cv2.CV_32F, 0, 1, ksize=ksize)

        grad_mag = np.sqrt(grad_x ** 2 + grad_y ** 2)

        # Create a mask: 1 where no edge, 0 where edge
        edge_mask = (grad_mag < threshold).astype(np.float32)

        # Apply mask
        depth_masked = depth * edge_mask
        batch_out.append(depth_masked)

    return np.stack(batch_out)


class PrepareForNet:
    def __init__(self, size, seed=123456, n_angles=4, max_angle_deg=45):
        self.size = size
        self.rng = random.Random(seed)
        # self.__eval_mode = eval_mode
        # self.__train_mode = not eval_mode

        self.n_angles = n_angles
        self.max_angle_deg = max_angle_deg
        self.lon_angles = np.linspace(0, max_angle_deg, n_angles)

    # def eval(self):
    #     self.__eval_mode = True
    #     self.__train_mode = False

    # def train(self):
    #     self.__train_mode = True
    #     self.__eval_mode = False

    def __call__(self, sample_with_key):

        sample = sample_with_key["data"]
        angle_idx = sample_with_key["angle_idx"]
        do_flip = sample_with_key["flip"]

        delta_lon = float(self.lon_angles[angle_idx])
        delta_lat = 0.0

        origin = sample['origin']
        sample.pop('origin')


        H, W = 14 * (self.size // 14), 14 * (self.size // 14)
        new_sample = {}

        for k, v in sample.items():
            if k == ' origin':
                continue

            if do_flip:
                # print(k,v.shape)
                if 'rgb'in k:
                    v = np.flip(v,axis=-2)
                else:
                    v = np.flip(v, axis=-1)

            # if 'rgb' in k: print(f'PrepareForNet v_max: {v.max()}')
                
            e = EquirectangularImage(v)
            
            # if os.environ.get('PLOT_DEBUG', False):
            #     import matplotlib.pyplot as plt
            #     plt.title(f'PrepareForNet: {k}')
            #     plt.imshow(e)
            #     plt.show()
            e.preprocess(shadow_angle=30 if 'p74' in origin else 0, delta_lat=delta_lat, delta_lon=delta_lon)
            
            # if os.environ.get('PLOT_DEBUG', False):
            #     plt.title(f'PrepareForNet: {k} - processed')
            #     plt.imshow(e)
            #     plt.show()
            e.attach_projection("gnomonic", x_points=W, y_points=H)
            
            faceset_a = np.array(e.to_gnomonic_face_set())

            e.attach_sampler("fibonacci", x_points=W, y_points=H, n_points=16)
            
            faceset_b = np.array(e.to_gnomonic_face_set())

            faceset = np.vstack([faceset_a, faceset_b])

            # print(f'[FACESET SHAPE]: {faceset.shape}')

            # if 'rgb' in k: print(f'PrepareForNet faceset_max: {faceset.max()}')

            if 'xyz' in k:
               u, v_ = np.meshgrid(
                   np.linspace(-1, 1, W),
                   np.linspace(-1, 1, H),
                   indexing='xy'
               )
               faceset = np.sqrt(faceset**2 / (u[None]**2 + v_[None]**2 + 1))
            #    faceset = zero_depth_edges_batch(faceset, ksize=3, threshold=.1)
            
            #if 'rgb' in k:
            #    faceset = faceset / 255.
            
            # if 'rgb' in k: print(f'PrepareForNet faceset_max: {faceset.max()}')

            # if os.environ.get('PLOT_DEBUG', False):
            import matplotlib.pyplot as plt
            for face in faceset:
                plt.title(f'[zeroed edges] PrepareForNet: {k}')
                plt.imshow(face)
                plt.show()
                break

            new_sample[k] = faceset.astype(np.float32)
            new_sample[k] = np.ascontiguousarray(new_sample[k])

        # Add metadata
        new_sample["angle_idx"] = angle_idx
        new_sample["flip"] = do_flip
        

        return new_sample
    
class _PrepareForNet:
    def __init__(self, size, seed=123456, eval_mode = False):
        self.size = size
        self.rng = random.Random(seed)
        self.__eval_mode = eval_mode
        self.__train_mode= not eval_mode

    
    def eval(self):
        self.__eval_mode = True
        self.__train_mode = False

    def train(self):
        self.__train_mode = True
        self.__eval_mode = False
        
    def __call__(self, sample):
        if self.__train_mode:
            delta_lon = self.rng.uniform(0, 45)
            delta_lat = 0.0
            do_flip = self.rng.random() < 0.5  # 50% chance to flip
            do_blur = self.rng.random() < 0.5  # 50% chance to flip


        else:
            delta_lon, delta_lat = (0.0, 0.0)
            do_flip = False
            do_blur = False

        new_sample = {}
        H, W = 14 * (self.size // 14), 14 * (self.size // 14)
        for k,v in sample.items():
            if do_flip:
                v = np.flip(v, axis=-1)
            #if (do_blur) & (self.__train_mode) & ('rgb' in k):
            #    blur_val = random.randint(5,15) #blur value random
            #    v = cv2.blur(v,(blur_val, blur_val))
                
            e = EquirectangularImage(v)
            e.preprocess(shadow_angle=30, delta_lat=delta_lat, delta_lon=delta_lon)
            
            e.attach_projection("gnomonic", x_points= W, y_points = H)
            faceset = np.array(e.to_gnomonic_face_set())

            if 'xyz' in k:

                u,v = np.meshgrid(
                    np.linspace(-1, 1, W),
                    np.linspace(-1, 1, H),
                    indexing='xy'
                )
    
                faceset = np.sqrt( faceset**2 / (u[None]**2 + v[None]**2 + 1))

            
            new_sample[k] = np.array(faceset).astype(np.float32)
            new_sample[k] = np.ascontiguousarray(new_sample[k])
        return new_sample
        

class NormalizeImage(object):
    """Normlize image by given mean and std.
    """
    __keyless__ = True
    def __init__(self, mean, std):
        self.__mean = mean
        self.__std = std

    def __call__(self, sample):
        sample["rgb_image"] = (sample["rgb_image"] / 255. - self.__mean) / self.__std

        return sample




_base_=[
       '../_base_/models/encoder_decoder/dino_vit_giant2_reg.dpt_raft.py',
       '../_base_/datasets/_data_base_.py',
       '../_base_/default_runtime.py',
       ]

import numpy as np
model=dict(
    decode_head=dict(
        type='RAFTDepthNormalDPT5',
        iters=8,
        n_downsample=2,
        detach=False,
    )
)


max_value = 200
# configs of the canonical space
data_basic=dict(
    canonical_space = dict(
        # img_size=(540, 960),
        focal_length=1000.0,
    ),
    depth_range=(0, 1),
    depth_normalize=(0.1, max_value),
    crop_size = (616, 1064),  # %28 = 0
     clip_depth_range=(0.1, 200),
    vit_size=(616,1064)
) 

batchsize_per_gpu = 1
thread_per_gpu = 1



from .SiLog import SilogLoss
from .WCEL import  WCELoss
from .VNL import VNLoss
from .Gradient import GradientLoss_Li, GradientLoss
from .Ranking import EdgeguidedRankingLoss, RankingLoss
from .Regularization import RegularizationLoss
from .SSIL import SSILoss
from .HDNL import HDNLoss
from .HDSNL import HDSNLoss
from .NormalRegression import EdgeguidedNormalLoss
from .depth_to_normal import Depth2Normal
from .photometric_loss_functions import PhotometricGeometricLoss
from .HDSNL_random import HDSNRandomLoss
from .HDNL_random import HDNRandomLoss
from .AdabinsLoss import AdabinsLoss
from .SkyRegularization import SkyRegularizationLoss
from .PWN_Planes import PWNPlanesLoss
from .L1 import L1Loss, L1DispLoss, L1InverseLoss
from .ConfidenceLoss import ConfidenceLoss
from .ScaleInvL1 import ScaleInvL1Loss
from .NormalBranchLoss import NormalBranchLoss, DeNoConsistencyLoss
from .GRUSequenceLoss import GRUSequenceLoss
from .ConfidenceGuideLoss import ConfidenceGuideLoss
from .ScaleAlignLoss import ScaleAlignLoss

__all__ = [
    'SilogLoss', 'WCELoss', 'VNLoss', 'GradientLoss_Li', 'GradientLoss', 'EdgeguidedRankingLoss',
    'RankingLoss', 'RegularizationLoss', 'SSILoss', 'HDNLoss', 'HDSNLoss', 'EdgeguidedNormalLoss', 'Depth2Normal',
    'PhotometricGeometricLoss', 'HDSNRandomLoss', 'HDNRandomLoss', 'AdabinsLoss', 'SkyRegularizationLoss',
    'PWNPlanesLoss', 'L1Loss',
    'ConfidenceLoss', 'ScaleInvL1Loss', 'L1DispLoss', 'NormalBranchLoss', 'L1InverseLoss', 'GRUSequenceLoss', 'ConfidenceGuideLoss', 'DeNoConsistencyLoss', 'ScaleAlignLoss'
]



#!/usr/bin/env python3
# Copyright (C) 2024-present Naver Corporation. All rights reserved.
# Licensed under CC BY-NC-SA 4.0 (non-commercial use only).
#
# --------------------------------------------------------
# Preprocessing code for the StaticThings3D dataset
# dataset at https://github.com/lmb-freiburg/robustmvd/blob/master/rmvd/data/README.md#staticthings3d
# 1) Download StaticThings3D in /path/to/StaticThings3D/
#    with the script at https://github.com/lmb-freiburg/robustmvd/blob/master/rmvd/data/scripts/download_staticthings3d.sh
#    --> depths.tar.bz2 frames_finalpass.tar.bz2 poses.tar.bz2 frames_cleanpass.tar.bz2 intrinsics.tar.bz2
# 2) unzip everything in the same /path/to/StaticThings3D/ directory
# 5) python datasets_preprocess/preprocess_staticthings3d.py --StaticThings3D_dir /path/to/tmp/StaticThings3D/
# --------------------------------------------------------
import os
import os.path as osp
import re
from tqdm import tqdm
import numpy as np
os.environ["OPENCV_IO_ENABLE_OPENEXR"] = "1"
import cv2

import path_to_root  # noqa
from dust3r.utils.parallel import parallel_threads
from dust3r.datasets.utils import cropping  # noqa


def get_parser():
    import argparse
    parser = argparse.ArgumentParser()
    parser.add_argument('--StaticThings3D_dir', required=True)
    parser.add_argument('--precomputed_pairs', required=True)
    parser.add_argument('--output_dir', default='data/staticthings3d_processed')
    return parser


def main(db_root, pairs_path, output_dir):
    all_scenes = _list_all_scenes(db_root)

    # crop images
    args = [(db_root, osp.join(split, subsplit, seq), camera, f'{n:04d}', output_dir)
            for split, subsplit, seq in all_scenes for camera in ['left', 'right'] for n in range(6, 16)]
    parallel_threads(load_crop_and_save, args, star_args=True, front_num=1)

    # verify that all images are there
    CAM = {b'l': 'left', b'r': 'right'}
    pairs = np.load(pairs_path)
    for scene, seq, cam1, im1, cam2, im2 in tqdm(pairs):
        seq_path = osp.join('TRAIN', scene.decode('ascii'), f'{seq:04d}')
        for cam, idx in [(CAM[cam1], im1), (CAM[cam2], im2)]:
            for ext in ['clean', 'final']:
                impath = osp.join(output_dir, seq_path, cam, f"{idx:04n}_{ext}.jpg")
                assert osp.isfile(impath), f'missing an image at {impath=}'

    print(f'>> Saved all data to {output_dir}!')


def load_crop_and_save(db_root, relpath_, camera, num, out_dir):
    relpath = osp.join(relpath_, camera, num)
    if osp.isfile(osp.join(out_dir, relpath + '.npz')):
        return
    os.makedirs(osp.join(out_dir, relpath_, camera), exist_ok=True)

    # load everything
    intrinsics_in = readFloat(osp.join(db_root, 'intrinsics', relpath_, num + '.float3'))
    cam2world = np.linalg.inv(readFloat(osp.join(db_root, 'poses', relpath + '.float3')))
    depthmap_in = readFloat(osp.join(db_root, 'depths', relpath + '.float3'))
    img_clean = cv2.cvtColor(cv2.imread(osp.join(db_root, 'frames_cleanpass',
                             relpath + '.png'), cv2.IMREAD_COLOR), cv2.COLOR_BGR2RGB)
    img_final = cv2.cvtColor(cv2.imread(osp.join(db_root, 'frames_finalpass',
                             relpath + '.png'), cv2.IMREAD_COLOR), cv2.COLOR_BGR2RGB)

    # do the crop
    assert img_clean.shape[:2] == (540, 960)
    assert img_final.shape[:2] == (540, 960)
    (clean_out, final_out), depthmap, intrinsics_out, R_in2out = _crop_image(
        intrinsics_in, (img_clean, img_final), depthmap_in, (512, 384))

    # write everything
    clean_out.save(osp.join(out_dir, relpath + '_clean.jpg'), quality=80)
    final_out.save(osp.join(out_dir, relpath + '_final.jpg'), quality=80)
    cv2.imwrite(osp.join(out_dir, relpath + '.exr'), depthmap)

    # New camera parameters
    cam2world[:3, :3] = cam2world[:3, :3] @ R_in2out.T
    np.savez(osp.join(out_dir, relpath + '.npz'), intrinsics=intrinsics_out, cam2world=cam2world)


def _crop_image(intrinsics_in, color_image_in, depthmap_in, resolution_out=(512, 512)):
    image, depthmap, intrinsics_out = cropping.rescale_image_depthmap(
        color_image_in, depthmap_in, intrinsics_in, resolution_out)
    R_in2out = np.eye(3)
    return image, depthmap, intrinsics_out, R_in2out


def _list_all_scenes(path):
    print('>> Listing all scenes')

    res = []
    for split in ['TRAIN']:
        for subsplit in 'ABC':
            for seq in os.listdir(osp.join(path, 'intrinsics', split, subsplit)):
                res.append((split, subsplit, seq))
    print(f'   (found ({len(res)}) scenes)')
    assert res, f'Did not find anything at {path=}'
    return res


def readFloat(name):
    with open(name, 'rb') as f:
        if (f.readline().decode("utf-8")) != 'float\n':
            raise Exception('float file %s did not contain <float> keyword' % name)

        dim = int(f.readline())

        dims = []
        count = 1
        for i in range(0, dim):
            d = int(f.readline())
            dims.append(d)
            count *= d

        dims = list(reversed(dims))
        data = np.fromfile(f, np.float32, count).reshape(dims)
    return data  # Hxw or CxHxW NxCxHxW


if __name__ == '__main__':
    parser = get_parser()
    args = parser.parse_args()
    main(args.StaticThings3D_dir, args.precomputed_pairs, args.output_dir)



### panorai_data_rewritten/utils/shape_manager.py ###
import numpy as np
from typing import Union

TensorOrArray = Union[np.ndarray]

class ShapeManager:
    """
    Utility class for managing and validating array shapes.

    - Converts inputs to NumPy arrays.
    - Used during loading, preprocessing, and conversion steps.
    """

    @staticmethod
    def to_numpy(data: TensorOrArray, dtype=np.float32) -> np.ndarray:
        """
        Converts arbitrary input to a NumPy array with optional dtype.

        Args:
            data (TensorOrArray): Input data.
            dtype (type, optional): Data type to convert to. Defaults to np.float32.

        Returns:
            np.ndarray: Standardized NumPy array.
        """
        return np.array(data, dtype=dtype)


import torch
import torch.nn.functional as F
import logging
import os
import os.path as osp
from mono.utils.avg_meter import MetricAverageMeter
from mono.utils.visualization import save_val_imgs, create_html, save_raw_imgs, save_normal_val_imgs
import cv2
from tqdm import tqdm
import numpy as np
from PIL import Image
import matplotlib.pyplot as plt

from mono.utils.unproj_pcd import reconstruct_pcd, save_point_cloud

def to_cuda(data: dict):
    for k, v in data.items():
        if isinstance(v, torch.Tensor):
            data[k] = v.cuda(non_blocking=True)
        if isinstance(v, list) and len(v)>=1 and isinstance(v[0], torch.Tensor):
            for i, l_i in enumerate(v):
                data[k][i] = l_i.cuda(non_blocking=True)
    return data

def align_scale(pred: torch.tensor, target: torch.tensor):
    mask = target > 0
    if torch.sum(mask) > 10:
        scale = torch.median(target[mask]) / (torch.median(pred[mask]) + 1e-8)
    else:
        scale = 1
    pred_scaled = pred * scale
    return pred_scaled, scale

def align_scale_shift(pred: torch.tensor, target: torch.tensor):
    mask = target > 0
    target_mask = target[mask].cpu().numpy()
    pred_mask = pred[mask].cpu().numpy()
    if torch.sum(mask) > 10:
        scale, shift = np.polyfit(pred_mask, target_mask, deg=1)
        if scale < 0:
            scale = torch.median(target[mask]) / (torch.median(pred[mask]) + 1e-8)
            shift = 0
    else:
        scale = 1
        shift = 0
    pred = pred * scale + shift
    return pred, scale

def align_scale_shift_numpy(pred: np.array, target: np.array):
    mask = target > 0
    target_mask = target[mask]
    pred_mask = pred[mask]
    if np.sum(mask) > 10:
        scale, shift = np.polyfit(pred_mask, target_mask, deg=1)
        if scale < 0:
            scale = np.median(target[mask]) / (np.median(pred[mask]) + 1e-8)
            shift = 0
    else:
        scale = 1
        shift = 0
    pred = pred * scale + shift
    return pred, scale


def build_camera_model(H : int, W : int, intrinsics : list) -> np.array:
    """
    Encode the camera intrinsic parameters (focal length and principle point) to a 4-channel map. 
    """
    fx, fy, u0, v0 = intrinsics
    f = (fx + fy) / 2.0
    # principle point location
    x_row = np.arange(0, W).astype(np.float32)
    x_row_center_norm = (x_row - u0) / W
    x_center = np.tile(x_row_center_norm, (H, 1)) # [H, W]

    y_col = np.arange(0, H).astype(np.float32) 
    y_col_center_norm = (y_col - v0) / H
    y_center = np.tile(y_col_center_norm, (W, 1)).T # [H, W]

    # FoV
    fov_x = np.arctan(x_center / (f / W))
    fov_y = np.arctan(y_center / (f / H))

    cam_model = np.stack([x_center, y_center, fov_x, fov_y], axis=2)
    return cam_model

def resize_for_input(image, output_shape, intrinsic, canonical_shape, to_canonical_ratio):
    """
    Resize the input.
    Resizing consists of two processed, i.e. 1) to the canonical space (adjust the camera model); 2) resize the image while the camera model holds. Thus the
    label will be scaled with the resize factor.
    """
    padding = [123.675, 116.28, 103.53]
    h, w, _ = image.shape
    resize_ratio_h = output_shape[0] / canonical_shape[0]
    resize_ratio_w = output_shape[1] / canonical_shape[1]
    to_scale_ratio = min(resize_ratio_h, resize_ratio_w)

    resize_ratio = to_canonical_ratio * to_scale_ratio

    reshape_h = int(resize_ratio * h)
    reshape_w = int(resize_ratio * w)

    pad_h = max(output_shape[0] - reshape_h, 0)
    pad_w = max(output_shape[1] - reshape_w, 0)
    pad_h_half = int(pad_h / 2)
    pad_w_half = int(pad_w / 2)

    # resize
    image = cv2.resize(image, dsize=(reshape_w, reshape_h), interpolation=cv2.INTER_LINEAR)
    # padding
    image = cv2.copyMakeBorder(
        image, 
        pad_h_half, 
        pad_h - pad_h_half, 
        pad_w_half, 
        pad_w - pad_w_half, 
        cv2.BORDER_CONSTANT, 
        value=padding)
    
    # Resize, adjust principle point
    intrinsic[2] = intrinsic[2] * to_scale_ratio
    intrinsic[3] = intrinsic[3] * to_scale_ratio

    cam_model = build_camera_model(reshape_h, reshape_w, intrinsic)
    cam_model = cv2.copyMakeBorder(
        cam_model, 
        pad_h_half, 
        pad_h - pad_h_half, 
        pad_w_half, 
        pad_w - pad_w_half, 
        cv2.BORDER_CONSTANT, 
        value=-1)

    pad=[pad_h_half, pad_h - pad_h_half, pad_w_half, pad_w - pad_w_half]
    label_scale_factor=1/to_scale_ratio
    return image, cam_model, pad, label_scale_factor


def get_prediction(
    model: torch.nn.Module,
    input: torch.tensor,
    cam_model: torch.tensor,
    pad_info: torch.tensor,
    scale_info: torch.tensor,
    gt_depth: torch.tensor,
    normalize_scale: float,
    ori_shape: list=[],
):

    data = dict(
        input=input,
        cam_model=cam_model,
    )
    pred_depth, confidence, output_dict = model.module.inference(data)

    return pred_depth, output_dict

def transform_test_data_scalecano(rgb, intrinsic, data_basic):
    """
    Pre-process the input for forwarding. Employ `label scale canonical transformation.'
        Args:
            rgb: input rgb image. [H, W, 3]
            intrinsic: camera intrinsic parameter, [fx, fy, u0, v0]
            data_basic: predefined canonical space in configs.
    """
    canonical_space = data_basic['canonical_space']
    forward_size = data_basic.crop_size
    mean = torch.tensor([123.675, 116.28, 103.53]).float()[:, None, None]
    std = torch.tensor([58.395, 57.12, 57.375]).float()[:, None, None]

    # BGR to RGB
    #rgb = cv2.cvtColor(rgb, cv2.COLOR_BGR2RGB)

    ori_h, ori_w, _ = rgb.shape
    ori_focal = (intrinsic[0] + intrinsic[1]) / 2
    canonical_focal = canonical_space['focal_length']

    cano_label_scale_ratio = canonical_focal / ori_focal

    canonical_intrinsic = [
        intrinsic[0] * cano_label_scale_ratio,
        intrinsic[1] * cano_label_scale_ratio,
        intrinsic[2],
        intrinsic[3],
    ]

    # resize
    rgb, cam_model, pad, resize_label_scale_ratio = resize_for_input(rgb, forward_size, canonical_intrinsic, [ori_h, ori_w], 1.0)

    # label scale factor
    label_scale_factor = cano_label_scale_ratio * resize_label_scale_ratio

    rgb = torch.from_numpy(rgb.transpose((2, 0, 1))).float()
    rgb = torch.div((rgb - mean), std)
    rgb = rgb.cuda()
    
    cam_model = torch.from_numpy(cam_model.transpose((2, 0, 1))).float()
    cam_model = cam_model[None, :, :, :].cuda()
    cam_model_stacks = [
        torch.nn.functional.interpolate(cam_model, size=(cam_model.shape[2]//i, cam_model.shape[3]//i), mode='bilinear', align_corners=False)
        for i in [2, 4, 8, 16, 32]
    ]
    return rgb, cam_model_stacks, pad, label_scale_factor

def do_scalecano_test_with_custom_data(
    model: torch.nn.Module,
    cfg: dict,
    test_data: list,
    logger: logging.RootLogger,
    is_distributed: bool = True,
    local_rank: int = 0,
    bs: int = 2,  # Batch size parameter
):

    show_dir = cfg.show_dir
    save_interval = 1
    save_imgs_dir = show_dir + '/vis'
    os.makedirs(save_imgs_dir, exist_ok=True)
    save_pcd_dir = show_dir + '/pcd'
    os.makedirs(save_pcd_dir, exist_ok=True)

    normalize_scale = cfg.data_basic.depth_range[1]
    dam = MetricAverageMeter(['abs_rel', 'rmse', 'silog', 'delta1', 'delta2', 'delta3'])
    dam_median = MetricAverageMeter(['abs_rel', 'rmse', 'silog', 'delta1', 'delta2', 'delta3'])
    dam_global = MetricAverageMeter(['abs_rel', 'rmse', 'silog', 'delta1', 'delta2', 'delta3'])
    
    # Process data in batches
    for i in tqdm(range(0, len(test_data), bs)):
        batch_data = test_data[i:i + bs]  # Extract batch
        rgb_inputs, pads, label_scale_factors, gt_depths, rgb_origins = [], [], [], [], []
        
        for an in batch_data:
            print(an['rgb'])
            rgb_origin = cv2.imread(an['rgb'])[:, :, ::-1].copy()
            rgb_origins.append(rgb_origin)
            gt_depth = None
            if an['depth'] is not None:
                gt_depth = cv2.imread(an['depth'], -1)
                gt_depth_scale = an['depth_scale']
                gt_depth = gt_depth / gt_depth_scale
            gt_depths.append(gt_depth)
            
            intrinsic = an['intrinsic']
            if intrinsic is None:
                intrinsic = [1000.0, 1000.0, rgb_origin.shape[1]/2, rgb_origin.shape[0]/2]
            
            rgb_input, _, pad, label_scale_factor = transform_test_data_scalecano(rgb_origin, intrinsic, cfg.data_basic)
            rgb_inputs.append(rgb_input)
            pads.append(pad)
            label_scale_factors.append(label_scale_factor)

        # Process the batch
        pred_depths, outputs = get_prediction(
            model=model,
            input=torch.stack(rgb_inputs),  # Stack inputs for batch processing
            cam_model=None,
            pad_info=pads,
            scale_info=None,
            gt_depth=None,
            normalize_scale=None,
        )
        
        for j, gt_depth in enumerate(gt_depths):
            normal_out = None
            if 'prediction_normal'  in outputs.keys():
                normal_out = outputs['prediction_normal'][j, :]
            
            postprocess_per_image(
                i*bs+j,
                pred_depths[j, :],
                gt_depth,
                intrinsic,
                rgb_origins[j],
                normal_out,
                pads[j],
                batch_data[j],
                dam,
                dam_median,
                dam_global,
                is_distributed,
                save_imgs_dir,
                save_pcd_dir,
                normalize_scale,
                label_scale_factors[j],
            )

    #if gt_depth_flag:
    if False:
        eval_error = dam.get_metrics()
        print('w/o match :', eval_error)

        eval_error_median = dam_median.get_metrics()
        print('median match :', eval_error_median)

        eval_error_global = dam_global.get_metrics()
        print('global match :', eval_error_global)
    else:
        print('missing gt_depth, only save visualizations...')


def postprocess_per_image(i, pred_depth, gt_depth, intrinsic, rgb_origin, normal_out, pad, an, dam, dam_median, dam_global, is_distributed, save_imgs_dir, save_pcd_dir, normalize_scale, scale_info):

    pred_depth = pred_depth.squeeze()
    pred_depth = pred_depth[pad[0] : pred_depth.shape[0] - pad[1], pad[2] : pred_depth.shape[1] - pad[3]]
    pred_depth = torch.nn.functional.interpolate(pred_depth[None, None, :, :], [rgb_origin.shape[0], rgb_origin.shape[1]], mode='bilinear').squeeze() # to original size
    pred_depth = pred_depth * normalize_scale / scale_info

    pred_depth = (pred_depth > 0) * (pred_depth < 300) * pred_depth
    if gt_depth is not None:

        pred_depth = torch.nn.functional.interpolate(pred_depth[None, None, :, :], (gt_depth.shape[0], gt_depth.shape[1]), mode='bilinear').squeeze() # to original size

        gt_depth = torch.from_numpy(gt_depth).cuda()

        pred_depth_median = pred_depth * gt_depth[gt_depth != 0].median() / pred_depth[gt_depth != 0].median()
        pred_global, _ = align_scale_shift(pred_depth, gt_depth)
        
        mask = (gt_depth > 1e-8)
        dam.update_metrics_gpu(pred_depth, gt_depth, mask, is_distributed)
        dam_median.update_metrics_gpu(pred_depth_median, gt_depth, mask, is_distributed)
        dam_global.update_metrics_gpu(pred_global, gt_depth, mask, is_distributed)
        print(gt_depth[gt_depth != 0].median() / pred_depth[gt_depth != 0].median(), )
    
    os.makedirs(osp.join(save_imgs_dir, an['folder']), exist_ok=True)
    rgb_torch = torch.from_numpy(rgb_origin).to(pred_depth.device).permute(2, 0, 1)
    mean = torch.tensor([123.675, 116.28, 103.53]).float()[:, None, None].to(rgb_torch.device)
    std = torch.tensor([58.395, 57.12, 57.375]).float()[:, None, None].to(rgb_torch.device)
    rgb_torch = torch.div((rgb_torch - mean), std)

    save_val_imgs(
        i,
        pred_depth,
        gt_depth if gt_depth is not None else torch.ones_like(pred_depth, device=pred_depth.device),
        rgb_torch,
        osp.join(an['folder'], an['filename']),
        save_imgs_dir,
    )
    #save_raw_imgs(pred_depth.detach().cpu().numpy(), rgb_torch, osp.join(an['folder'], an['filename']), save_imgs_dir, 1000.0)

    # pcd
    pred_depth = pred_depth.detach().cpu().numpy()
    #pcd = reconstruct_pcd(pred_depth, intrinsic[0], intrinsic[1], intrinsic[2], intrinsic[3])
    #os.makedirs(osp.join(save_pcd_dir, an['folder']), exist_ok=True)
    #save_point_cloud(pcd.reshape((-1, 3)), rgb_origin.reshape(-1, 3), osp.join(save_pcd_dir, an['folder'], an['filename'][:-4]+'.ply'))

    if an['intrinsic'] == None:
        #for r in [0.9, 1.0, 1.1]:
        for r in [1.0]:
            #for f in [600, 800, 1000, 1250, 1500]:
            for f in [1000]:
                pcd = reconstruct_pcd(pred_depth, f * r, f * (2-r), intrinsic[2], intrinsic[3])
                fstr = '_fx_' + str(int(f * r)) + '_fy_' + str(int(f * (2-r)))
                os.makedirs(osp.join(save_pcd_dir, an['folder']), exist_ok=True)
                save_point_cloud(pcd.reshape((-1, 3)), rgb_origin.reshape(-1, 3), osp.join(save_pcd_dir, an['folder'], an['filename'][:-4] + fstr +'.ply'))

    if normal_out is not None:
        pred_normal = normal_out[:3, :, :] # (3, H, W)
        H, W = pred_normal.shape[1:]
        pred_normal = pred_normal[ :, pad[0]:H-pad[1], pad[2]:W-pad[3]]

        pred_normal = torch.nn.functional.interpolate(pred_normal[None, :], size=[rgb_origin.shape[0], rgb_origin.shape[1]], mode='bilinear', align_corners=True).squeeze()

        gt_normal = None
        #if gt_normal_flag:
        if False:
            pred_normal = torch.nn.functional.interpolate(pred_normal, size=gt_normal.shape[2:], mode='bilinear', align_corners=True)    
            gt_normal = cv2.imread(norm_path)
            gt_normal = cv2.cvtColor(gt_normal, cv2.COLOR_BGR2RGB) 
            gt_normal = np.array(gt_normal).astype(np.uint8)
            gt_normal = ((gt_normal.astype(np.float32) / 255.0) * 2.0) - 1.0
            norm_valid_mask = (np.linalg.norm(gt_normal, axis=2, keepdims=True) > 0.5)
            gt_normal = gt_normal * norm_valid_mask               
            gt_normal_mask = ~torch.all(gt_normal == 0, dim=1, keepdim=True)
            dam.update_normal_metrics_gpu(pred_normal, gt_normal, gt_normal_mask, cfg.distributed)# save valiad normal

        save_normal_val_imgs(iter, 
                            pred_normal, 
                            gt_normal if gt_normal is not None else torch.ones_like(pred_normal, device=pred_normal.device),
                            rgb_torch, # data['input'], 
                            osp.join(an['folder'], 'normal_'+an['filename']), 
                            save_imgs_dir,
                            )




# MIT License

# Copyright (c) 2022 Intelligent Systems Lab Org

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# File author: Shariq Farooq Bhat

import torch
import torch.nn as nn


def log_binom(n, k, eps=1e-7):
    """ log(nCk) using stirling approximation """
    n = n + eps
    k = k + eps
    return n * torch.log(n) - k * torch.log(k) - (n-k) * torch.log(n-k+eps)


class LogBinomial(nn.Module):
    def __init__(self, n_classes=256, act=torch.softmax):
        """Compute log binomial distribution for n_classes

        Args:
            n_classes (int, optional): number of output classes. Defaults to 256.
        """
        super().__init__()
        self.K = n_classes
        self.act = act
        self.register_buffer('k_idx', torch.arange(
            0, n_classes).view(1, -1, 1, 1))
        self.register_buffer('K_minus_1', torch.Tensor(
            [self.K-1]).view(1, -1, 1, 1))

    def forward(self, x, t=1., eps=1e-4):
        """Compute log binomial distribution for x

        Args:
            x (torch.Tensor - NCHW): probabilities
            t (float, torch.Tensor - NCHW, optional): Temperature of distribution. Defaults to 1..
            eps (float, optional): Small number for numerical stability. Defaults to 1e-4.

        Returns:
            torch.Tensor -NCHW: log binomial distribution logbinomial(p;t)
        """
        if x.ndim == 3:
            x = x.unsqueeze(1)  # make it nchw

        one_minus_x = torch.clamp(1 - x, eps, 1)
        x = torch.clamp(x, eps, 1)
        y = log_binom(self.K_minus_1, self.k_idx) + self.k_idx * \
            torch.log(x) + (self.K - 1 - self.k_idx) * torch.log(one_minus_x)
        return self.act(y/t, dim=1)


class ConditionalLogBinomial(nn.Module):
    def __init__(self, in_features, condition_dim, n_classes=256, bottleneck_factor=2, p_eps=1e-4, max_temp=50, min_temp=1e-7, act=torch.softmax):
        """Conditional Log Binomial distribution

        Args:
            in_features (int): number of input channels in main feature
            condition_dim (int): number of input channels in condition feature
            n_classes (int, optional): Number of classes. Defaults to 256.
            bottleneck_factor (int, optional): Hidden dim factor. Defaults to 2.
            p_eps (float, optional): small eps value. Defaults to 1e-4.
            max_temp (float, optional): Maximum temperature of output distribution. Defaults to 50.
            min_temp (float, optional): Minimum temperature of output distribution. Defaults to 1e-7.
        """
        super().__init__()
        self.p_eps = p_eps
        self.max_temp = max_temp
        self.min_temp = min_temp
        self.log_binomial_transform = LogBinomial(n_classes, act=act)
        bottleneck = (in_features + condition_dim) // bottleneck_factor
        self.mlp = nn.Sequential(
            nn.Conv2d(in_features + condition_dim, bottleneck,
                      kernel_size=1, stride=1, padding=0),
            nn.GELU(),
            # 2 for p linear norm, 2 for t linear norm
            nn.Conv2d(bottleneck, 2+2, kernel_size=1, stride=1, padding=0),
            nn.Softplus()
        )

    def forward(self, x, cond):
        """Forward pass

        Args:
            x (torch.Tensor - NCHW): Main feature
            cond (torch.Tensor - NCHW): condition feature

        Returns:
            torch.Tensor: Output log binomial distribution
        """
        pt = self.mlp(torch.concat((x, cond), dim=1))
        p, t = pt[:, :2, ...], pt[:, 2:, ...]

        p = p + self.p_eps
        p = p[:, 0, ...] / (p[:, 0, ...] + p[:, 1, ...])

        t = t + self.p_eps
        t = t[:, 0, ...] / (t[:, 0, ...] + t[:, 1, ...])
        t = t.unsqueeze(1)
        t = (self.max_temp - self.min_temp) * t + self.min_temp

        return self.log_binomial_transform(p, t)



from ...data.gnomonic_image import GnomonicFace as GnomonicFace
from ..data import PCD as PCD
from ..handler import PCDHandler as PCDHandler
from .base_blender import BaseBlender as BaseBlender
from _typeshed import Incomplete

class KDTreeBlender(BaseBlender):
    merge_radius: Incomplete
    def __init__(self, merge_radius: float = 0.05, min_radius: float = 0.0, max_radius: float = 20.0, **kwargs) -> None: ...
    def process_faceset(self, faceset, model, grad_threshold: float = 0.1, feather_exp: float = 1.0): ...



# Copyright (C) 2024-present Naver Corporation. All rights reserved.
# Licensed under CC BY-NC-SA 4.0 (non-commercial use only).
#
# --------------------------------------------------------
# Dummy optimizer for visualizing pairs
# --------------------------------------------------------
import numpy as np
import torch
import torch.nn as nn
import cv2

from dust3r.cloud_opt.base_opt import BasePCOptimizer
from dust3r.utils.geometry import inv, geotrf, depthmap_to_absolute_camera_coordinates
from dust3r.cloud_opt.commons import edge_str
from dust3r.post_process import estimate_focal_knowing_depth


class PairViewer (BasePCOptimizer):
    """
    This a Dummy Optimizer.
    To use only when the goal is to visualize the results for a pair of images (with is_symmetrized)
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        assert self.is_symmetrized and self.n_edges == 2
        self.has_im_poses = True

        # compute all parameters directly from raw input
        self.focals = []
        self.pp = []
        rel_poses = []
        confs = []
        for i in range(self.n_imgs):
            conf = float(self.conf_i[edge_str(i, 1-i)].mean() * self.conf_j[edge_str(i, 1-i)].mean())
            if self.verbose:
                print(f'  - {conf=:.3} for edge {i}-{1-i}')
            confs.append(conf)

            H, W = self.imshapes[i]
            pts3d = self.pred_i[edge_str(i, 1-i)]
            pp = torch.tensor((W/2, H/2))
            focal = float(estimate_focal_knowing_depth(pts3d[None], pp, focal_mode='weiszfeld'))
            self.focals.append(focal)
            self.pp.append(pp)

            # estimate the pose of pts1 in image 2
            pixels = np.mgrid[:W, :H].T.astype(np.float32)
            pts3d = self.pred_j[edge_str(1-i, i)].numpy()
            assert pts3d.shape[:2] == (H, W)
            msk = self.get_masks()[i].numpy()
            K = np.float32([(focal, 0, pp[0]), (0, focal, pp[1]), (0, 0, 1)])

            try:
                res = cv2.solvePnPRansac(pts3d[msk], pixels[msk], K, None,
                                         iterationsCount=100, reprojectionError=5, flags=cv2.SOLVEPNP_SQPNP)
                success, R, T, inliers = res
                assert success

                R = cv2.Rodrigues(R)[0]  # world to cam
                pose = inv(np.r_[np.c_[R, T], [(0, 0, 0, 1)]])  # cam to world
            except:
                pose = np.eye(4)
            rel_poses.append(torch.from_numpy(pose.astype(np.float32)))

        # let's use the pair with the most confidence
        if confs[0] > confs[1]:
            # ptcloud is expressed in camera1
            self.im_poses = [torch.eye(4), rel_poses[1]]  # I, cam2-to-cam1
            self.depth = [self.pred_i['0_1'][..., 2], geotrf(inv(rel_poses[1]), self.pred_j['0_1'])[..., 2]]
        else:
            # ptcloud is expressed in camera2
            self.im_poses = [rel_poses[0], torch.eye(4)]  # I, cam1-to-cam2
            self.depth = [geotrf(inv(rel_poses[0]), self.pred_j['1_0'])[..., 2], self.pred_i['1_0'][..., 2]]

        self.im_poses = nn.Parameter(torch.stack(self.im_poses, dim=0), requires_grad=False)
        self.focals = nn.Parameter(torch.tensor(self.focals), requires_grad=False)
        self.pp = nn.Parameter(torch.stack(self.pp, dim=0), requires_grad=False)
        self.depth = nn.ParameterList(self.depth)
        for p in self.parameters():
            p.requires_grad = False

    def _set_depthmap(self, idx, depth, force=False):
        if self.verbose:
            print('_set_depthmap is ignored in PairViewer')
        return

    def get_depthmaps(self, raw=False):
        depth = [d.to(self.device) for d in self.depth]
        return depth

    def _set_focal(self, idx, focal, force=False):
        self.focals[idx] = focal

    def get_focals(self):
        return self.focals

    def get_known_focal_mask(self):
        return torch.tensor([not (p.requires_grad) for p in self.focals])

    def get_principal_points(self):
        return self.pp

    def get_intrinsics(self):
        focals = self.get_focals()
        pps = self.get_principal_points()
        K = torch.zeros((len(focals), 3, 3), device=self.device)
        for i in range(len(focals)):
            K[i, 0, 0] = K[i, 1, 1] = focals[i]
            K[i, :2, 2] = pps[i]
            K[i, 2, 2] = 1
        return K

    def get_im_poses(self):
        return self.im_poses

    def depth_to_pts3d(self):
        pts3d = []
        for d, intrinsics, im_pose in zip(self.depth, self.get_intrinsics(), self.get_im_poses()):
            pts, _ = depthmap_to_absolute_camera_coordinates(d.cpu().numpy(),
                                                             intrinsics.cpu().numpy(),
                                                             im_pose.cpu().numpy())
            pts3d.append(torch.from_numpy(pts).to(device=self.device))
        return pts3d

    def forward(self):
        return float('nan')



import torch
import torch.nn as nn

EPSILON = 1e-6
"""
  # @Zhengqi Li version.  
  def GradientLoss(self, log_prediction_d, mask, log_gt):
        log_d_diff = log_prediction_d - log_gt

        v_gradient = torch.abs(log_d_diff[:, :-2, :] - log_d_diff[:, 2:, :])
        v_mask = torch.mul(mask[:, :-2, :], mask[:, 2:, :])
        v_gradient = torch.mul(v_gradient, v_mask)

        h_gradient = torch.abs(log_d_diff[:, :, :-2] - log_d_diff[:, :, 2:])
        h_mask = torch.mul(mask[:, :, :-2], mask[:, :, 2:])
        h_gradient = torch.mul(h_gradient, h_mask)

        N = torch.sum(h_mask) + torch.sum(v_mask) + EPSILON

        gradient_loss = torch.sum(h_gradient) + torch.sum(v_gradient)
        gradient_loss = gradient_loss / N

        return gradient_loss
"""
def gradient_log_loss(log_prediction_d, log_gt, mask):
    log_d_diff = log_prediction_d - log_gt

    v_gradient = torch.abs(log_d_diff[:, :, :-2, :] - log_d_diff[:, :, 2:, :])
    v_mask = torch.mul(mask[:, :, :-2, :], mask[:, :, 2:, :])
    v_gradient = torch.mul(v_gradient, v_mask)

    h_gradient = torch.abs(log_d_diff[:, :, :, :-2] - log_d_diff[:, :, :, 2:])
    h_mask = torch.mul(mask[:, :, :, :-2], mask[:, :, :, 2:])
    h_gradient = torch.mul(h_gradient, h_mask)

    N = torch.sum(h_mask) + torch.sum(v_mask) + EPSILON

    gradient_loss = torch.sum(h_gradient) + torch.sum(v_gradient)
    gradient_loss = gradient_loss / N

    return gradient_loss

class GradientLoss_Li(nn.Module):
    def __init__(self, scale_num=1, loss_weight=1, data_type = ['lidar', 'stereo'], **kwargs):
        super(GradientLoss_Li, self).__init__()
        self.__scales = scale_num
        self.loss_weight = loss_weight
        self.data_type = data_type
        self.eps = 1e-6

    def forward(self, prediction, target, mask, **kwargs):
        total = 0
        target_trans = target + (~mask) * 100
        pred_log = torch.log(prediction)
        gt_log = torch.log(target_trans)
        for scale in range(self.__scales):
            step = pow(2, scale)
            
            total += gradient_log_loss(pred_log[:, ::step, ::step], gt_log[:, ::step, ::step], mask[:, ::step, ::step])
        loss = total / self.__scales
        if torch.isnan(loss).item() | torch.isinf(loss).item():
            raise RuntimeError(f'VNL error, {loss}')
        return loss * self.loss_weight
  
######################################################
# Multi-scale gradient matching loss, @Ke Xian implementation.
#####################################################
def gradient_loss(prediction, target, mask):
    M = torch.sum(mask, (1, 2))

    diff = prediction - target
    diff = torch.mul(mask, diff)

    grad_x = torch.abs(diff[:, :, 1:] - diff[:, :, :-1])
    mask_x = torch.mul(mask[:, :, 1:], mask[:, :, :-1])
    grad_x = torch.mul(mask_x, grad_x)

    grad_y = torch.abs(diff[:, 1:, :] - diff[:, :-1, :])
    mask_y = torch.mul(mask[:, 1:, :], mask[:, :-1, :])
    grad_y = torch.mul(mask_y, grad_y)

    image_loss = torch.sum(grad_x, (1, 2)) + torch.sum(grad_y, (1, 2))
    valid = M.nonzero()
    if image_loss[valid].numel() > 0:
        image_loss[valid] = image_loss[valid] / M[valid]
        loss = torch.mean(image_loss)
    else:
        loss = 0 * torch.sum(prediction)

    return loss


class GradientLoss(nn.Module):
    def __init__(self, scale_num=4, loss_weight=1, **kwargs):
        super(GradientLoss, self).__init__()
        self.__scales = scale_num
        self.loss_weight = loss_weight
    def forward(self, prediction, target, mask, **kwargs):
        total = 0
        for scale in range(self.__scales):
            step = pow(2, scale)
            total += gradient_loss(prediction[:, ::step, ::step], target[:, ::step, ::step], mask[:, ::step, ::step])
         
        return total * self.loss_weight


if __name__ == '__main__':
    import numpy as np
    gradient = GradientLoss_Li(4)

    pred_depth = np.random.random([2, 1, 480, 640])
    gt_depth = np.ones_like(pred_depth) * (-1) #np.random.random([2, 1, 480, 640]) - 0.5 #
    #gt_depth = np.abs(gt_depth)
    intrinsic = [[100, 100, 200, 200], [200, 200, 300, 300]]

    pred = torch.from_numpy(pred_depth).cuda()
    gt = torch.from_numpy(gt_depth).cuda()
    mask = gt > 0

    loss = gradient(gt, gt, mask)
    print(loss)


import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F

class SkyRegularizationLoss(nn.Module):
    """
    Enforce losses on pixels without any gts.
    """
    def __init__(self, loss_weight=0.1, data_type=['sfm', 'stereo', 'lidar', 'denselidar', 'denselidar_nometric', 'denselidar_syn'], sky_id=142, sample_ratio=0.4, regress_value=1.8, normal_regress=None, normal_weight=1.0, **kwargs):
        super(SkyRegularizationLoss, self).__init__()
        self.loss_weight = loss_weight
        self.data_type = data_type
        self.sky_id = sky_id
        self.sample_ratio = sample_ratio
        self.eps = 1e-6
        self.regress_value = regress_value
        self.normal_regress = normal_regress
        self.normal_weight = normal_weight
    
    def loss1(self, pred_sky):
        loss = 1/ torch.exp((torch.sum(pred_sky) / (pred_sky.numel() + self.eps)))
        return loss

    def loss2(self, pred_sky):
        loss = torch.sum(torch.abs(pred_sky - self.regress_value)) / (pred_sky.numel() + self.eps)
        return loss

    def loss_norm(self, pred_norm, sky_mask):
        sky_norm = torch.FloatTensor(self.normal_regress).cuda()
        sky_norm = sky_norm.unsqueeze(0).unsqueeze(2).unsqueeze(3)
        dot = torch.cosine_similarity(pred_norm[:, :3, :, :].clone(), sky_norm, dim=1)

        sky_mask_float = sky_mask.float().squeeze()
        valid_mask = sky_mask_float \
                        * (dot.detach() < 0.999).float() \
                        * (dot.detach() > -0.999).float() 

        al = (1 - dot) * valid_mask
        loss = torch.sum(al) / (torch.sum(sky_mask_float) + self.eps)
        return loss

    def forward(self, prediction, target, prediction_normal=None, mask=None, sem_mask=None,  **kwargs):
        sky_mask = sem_mask == self.sky_id
        pred_sky = prediction[sky_mask]
        pred_sky_numel = pred_sky.numel()

        if pred_sky.numel() > 50:
            samples = np.random.choice(pred_sky_numel, int(pred_sky_numel*self.sample_ratio), replace=False)
        
        if pred_sky.numel() > 0:
            #loss = - torch.sum(pred_wo_gt) / (pred_wo_gt.numel() + 1e-8)
            loss = self.loss2(pred_sky)

            if (prediction_normal != None) and (self.normal_regress != None):
                loss_normal = self.loss_norm(prediction_normal, sky_mask)
                loss = loss + loss_normal * self.normal_weight

        else:
            loss = torch.sum(prediction) * 0
        if torch.isnan(loss).item() | torch.isinf(loss).item():
            loss = torch.sum(prediction) * 0
            print(f'SkyRegularization NAN error, {loss}')    
        #    raise RuntimeError(f'Sky Loss error, {loss}')    
        
        return loss * self.loss_weight

if __name__ == '__main__':
    import cv2
    sky = SkyRegularizationLoss()
    pred_depth = np.random.random([2, 1, 480, 640])
    gt_depth = np.zeros_like(pred_depth) #np.random.random([2, 1, 480, 640])
    intrinsic = [[[100, 0, 200], [0, 100, 200], [0, 0, 1]], [[100, 0, 200], [0, 100, 200], [0, 0, 1]],]
    gt_depth = torch.tensor(np.array(gt_depth, np.float32)).cuda()
    pred_depth = torch.tensor(np.array(pred_depth, np.float32)).cuda()
    intrinsic = torch.tensor(np.array(intrinsic, np.float32)).cuda()
    mask = gt_depth > 0
    loss1 = sky(pred_depth, gt_depth, mask, mask, intrinsic)
    print(loss1)


import numpy as np
from _typeshed import Incomplete

class ArrayLikeData:
    def __array__(self, dtype: Incomplete | None = None, copy: bool = True) -> np.ndarray: ...
    def __array_ufunc__(self, ufunc, method, *inputs, **kwargs): ...
    def __getitem__(self, index) -> np.ndarray | ArrayLikeData: ...
    def __setitem__(self, index, value) -> None: ...
    @property
    def shape(self) -> tuple[int, ...]: ...



from typing import Tuple, Any
import numpy as np
import logging
import math
from ...utils.exceptions import TransformationError, ConfigurationError
from ..base.transform import BaseCoordinateTransformer

logger = logging.getLogger('spherical_projections.gnomonic_projection.gnomonic.transform')

class GnomonicTransformer(BaseCoordinateTransformer):
    """
    Transformer for Gnomonic Projection.
    Converts geographic (lat, lon) to image coordinates on the Gnomonic plane.
    """

    def __init__(self, config: Any):
        logger.debug("Initializing GnomonicTransformer.")
        required_attrs = ["lon_min", "lon_max", "lat_min", "lat_max", "fov_deg", "R", "x_points", "y_points"]
        missing = [attr for attr in required_attrs if not hasattr(config, attr)]
        if missing:
            raise ConfigurationError(f"Missing required config attributes: {', '.join(missing)}")
        self.config = config
        logger.info("GnomonicTransformer initialized.")

    def _compute_image_coords(self, values: np.ndarray, min_val: float, max_val: float, size: int) -> np.ndarray:
        normalized = (values - min_val) / (max_val - min_val) * (size - 1)
        logger.debug("Computed normalized image coordinates.")
        return normalized

    def spherical_to_image_coords(self, lat: np.ndarray, lon: np.ndarray, shape: Tuple[int, int]) -> Tuple[np.ndarray, np.ndarray]:
        logger.debug("Mapping spherical coordinates to image coordinates.")
        H, W = shape
        # Clamp extreme values if needed
        lon[lon > 180] = -360 + lon[lon > 180]
        lon[lon < -180] = 180 + (lon[lon < -180] + 180)
        lat[lat > 90] = -180 + lat[lat > 90]
        map_x = self._compute_image_coords(lon, self.config.lon_min, self.config.lon_max, W)
        map_y = self._compute_image_coords(lat, self.config.lat_max, self.config.lat_min, H)
        return map_x, map_y

    def projection_to_image_coords(self, x: np.ndarray, y: np.ndarray, config: Any) -> Tuple[np.ndarray, np.ndarray]:
        logger.debug("Mapping projection coordinates to image coordinates.")
        half_fov_rad = (config.fov_deg / 2) * math.pi / 180.0
        x_max = math.tan(half_fov_rad) * config.R
        y_max = math.tan(half_fov_rad) * config.R
        x_min, y_min = -x_max, -y_max
        map_x = self._compute_image_coords(x, x_min, x_max, config.x_points)
        map_y = self._compute_image_coords(y, y_max, y_min, config.y_points)
        return map_x, map_y



from .registry import ConfigRegistry as ConfigRegistry
from _typeshed import Incomplete
from typing import Any

logger: Incomplete

class ConfigManager:
    @classmethod
    def create(cls, name: str, **kwargs) -> Any: ...
    @classmethod
    def get(cls, name: str) -> Any: ...
    @classmethod
    def modify_config(cls, name: str, **kwargs): ...
    @classmethod
    def get_config_parameters(cls, name: str) -> dict[str, Any]: ...
    @classmethod
    def describe_config(cls, name: str, output_format: str = 'pretty') -> None: ...
    @classmethod
    def reset(cls, name: str | None = None): ...
    @classmethod
    def get_all_configs(cls) -> dict[str, Any]: ...



import torch
import torch.nn as nn

class HDNLoss(nn.Module):
    """
    Hieratical depth normalization loss.
    loss = MAE((d-median(d)/s - (d'-median(d'))/s'), s = mean(d- median(d))
    """
    def __init__(self, loss_weight=1, grid=3, data_type=['sfm', 'stereo', 'lidar'], **kwargs):
        super(HDNLoss, self).__init__()
        self.loss_weight = loss_weight
        self.grid = grid
        self.data_type = data_type
    
    def get_hierachy_masks(self, grid, depth_gt, mask_valid):

        batch_map_grid = []
        for mask_index in range(depth_gt.shape[0]):
            depth_map = depth_gt[mask_index]
            valid_map = mask_valid[mask_index]

            # print (depth_map[valid_map].view(-1).shape)
            if depth_map[valid_map].numel() == 0:
                map_grid_list = [valid_map for _ in range(2 ** (grid) - 1)]
            else:
                valid_values = depth_map[valid_map]

                max_d = valid_values.max()
                min_d = valid_values.min()

                anchor_power = [(1 / 2) ** (i) for i in range(grid)]
                anchor_power.reverse()

                map_grid_list = []
                for anchor in anchor_power:
                    # range
                    for i in range(int(1 / anchor)):
                        mask_new = (depth_map >= min_d + (max_d - min_d) * i * anchor) & (
                                    depth_map < min_d + (max_d - min_d) * (i + 1) * anchor+1e-30)
                        # print (f'[{i*anchor},{(i+1)*anchor}]')
                        mask_new = mask_new & valid_map
                        map_grid_list.append(mask_new)
            map_grid_list = torch.stack(map_grid_list, dim=0)
            batch_map_grid.append(map_grid_list)
        batch_map_grid = torch.stack(batch_map_grid, dim=1)
        return batch_map_grid
    
    def ssi_mae(self, prediction, target, mask_valid):
        B, C, H, W = target.shape
        prediction_nan = prediction.clone()
        target_nan = target.clone()
        prediction_nan[~mask_valid] = float('nan')
        target_nan[~mask_valid] = float('nan')

        valid_pixs = mask_valid.reshape((B, C,-1)).sum(dim=2, keepdims=True) + 1e-10
        valid_pixs = valid_pixs[:, :, :, None]

        gt_median = target_nan.reshape((B, C,-1)).nanmedian(2, keepdims=True)[0].unsqueeze(-1) # [b,c,h,w]
        gt_median[torch.isnan(gt_median)] = 0
        gt_diff = (torch.abs(target - gt_median) * mask_valid).reshape((B, C, -1))
        gt_s = gt_diff.sum(dim=2)[:, :, None, None] / valid_pixs
        gt_trans = (target - gt_median) / (gt_s + 1e-8)

        pred_median = prediction_nan.reshape((B, C,-1)).nanmedian(2, keepdims=True)[0].unsqueeze(-1) # [b,c,h,w]
        pred_median[torch.isnan(pred_median)] = 0
        pred_diff = (torch.abs(prediction - pred_median) * mask_valid).reshape((B, C, -1))
        pred_s = pred_diff.sum(dim=2)[:, :, None, None] / valid_pixs
        pred_trans = (prediction - pred_median) / (pred_s + 1e-8)

        loss = torch.sum(torch.abs(gt_trans - pred_trans)*mask_valid) / (torch.sum(mask_valid) + 1e-8)
        return pred_trans, gt_trans, loss

    def forward(self, prediction, target, mask=None, **kwargs):
        """
        Calculate loss.
        """
        B, C, H, W = target.shape
        hierachy_masks = self.get_hierachy_masks(self.grid, target, mask)
        hierachy_masks_shape = hierachy_masks.reshape(-1, C, H, W)    
        prediction_hie = prediction.unsqueeze(0).repeat(hierachy_masks.shape[0], 1, 1, 1, 1).reshape(-1, C, H, W)     

        target_hie = target.unsqueeze(0).repeat(hierachy_masks.shape[0], 1, 1, 1, 1).reshape(-1, C, H, W)

        #_, _, loss = self.ssi_mae(prediction, target, mask)
        _, _, loss = self.ssi_mae(prediction_hie, target_hie, hierachy_masks_shape)
        return loss * self.loss_weight
    
if __name__ == '__main__':
    ssil = HDNLoss()
    pred = torch.rand((2, 1, 256, 256)).cuda()
    gt = torch.rand((2, 1, 256, 256)).cuda()#torch.zeros_like(pred).cuda() #
    gt[:, :, 100:256, 0:100] = -1
    mask = gt > 0
    out = ssil(pred, gt, mask)
    print(out)
